# coding=utf-8
import Augmentor
import logging
import numpy as np
import os
from PIL import Image
import shutil
import tensorflow as tf
from tensorflow_hub.tools.make_image_classifier import (
    make_image_classifier as retrain_script,
)

import mammut
from mammut import BASE_MODEL_VISON_DIR

log = logging.getLogger(__name__)


# retrain_script = make_image_classifier.make_image_classifier


def set_dir_as_category(directory):
    images = os.listdir(directory)
    for index, image in enumerate(images):
        image_category = ".".join(image.split(".")[:-1])
        new_path = directory + "/" + image_category
        if not os.path.isdir(new_path):
            src = directory + "/" + image
            dst = directory + "/" + image_category + "/" + image
            if not os.path.exists(new_path) and os.path.isfile(src):
                os.makedirs(new_path)
            if os.path.isfile(src):
                shutil.move(src, dst)


def augmentate_images(directory, samples):
    image_categories = os.listdir(directory)
    for category in image_categories:
        # print(directory + category + '/')
        p = Augmentor.Pipeline(directory + category + "/", "")
        p.rotate(probability=0.7, max_left_rotation=10, max_right_rotation=10)
        p.zoom(probability=0.5, min_factor=1.1, max_factor=1.5)
        p.flip_left_right(probability=0.5)
        p.sample(samples)


def persist_retrained_model(retrained_graph_folder, labels):
    # obtener el current model name
    identifier = get_time_stamp_identifier(retrained_graph_folder, False)
    model_path, dir_base_name = (
        os.path.split(retrained_graph_folder)
        if os.path.split(retrained_graph_folder)[1] != ""
        else os.path.split(os.path.split(retrained_graph_folder)[0])
    )
    saved_model_dir = model_path + "/" + dir_base_name + identifier
    labels_output_file = saved_model_dir + "/" + "labels_output_file.txt"
    if os.path.exists(saved_model_dir):
        shutil.copy(labels, labels_output_file)
        shutil.move(saved_model_dir, retrained_graph_folder)


def retrain(
    image_dir,
    summaries_dir="/tmp/retrain_logs",
    tflite_output_file="/tmp/tflite_output_file.tflite",
    labels_output_file="/tmp/labels_output_file.txt",
    saved_model_dir="",
    train_epochs=10,
    should_create_new=True,
):
    # mod timestamp
    # si recibe crear_nuevo creamos uno nuevo
    # verificamos si hay mas que el maximo de modelos permitidos y borrmaos el mas viejo
    # sino devolvemos el mas nuevo/current
    identifier = get_time_stamp_identifier(saved_model_dir, should_create_new)
    model_path, dir_base_name = (
        os.path.split(saved_model_dir)
        if os.path.split(saved_model_dir)[1] != ""
        else os.path.split(os.path.split(saved_model_dir)[0])
    )
    saved_model_dir = model_path + "/" + dir_base_name + identifier
    print(saved_model_dir)

    if image_dir is not None:
        retrain_script.FLAGS.image_dir = image_dir
    # no available in tensorflow-hub 0.8 but looks like after it will (check master)
    # if summaries_dir is not None:
    #     retrain_script.FLAGS.summaries_dir = summaries_dir
    if tflite_output_file is not None:
        retrain_script.FLAGS.tflite_output_file = tflite_output_file
    if labels_output_file is not None:
        retrain_script.FLAGS.labels_output_file = labels_output_file
    if saved_model_dir is not None:
        retrain_script.FLAGS.saved_model_dir = saved_model_dir
    if train_epochs is not None:
        retrain_script.FLAGS.train_epochs = train_epochs
    for flag_key in retrain_script.FLAGS:
        flag = retrain_script.FLAGS[flag_key]
        flag.present += 1
    retrain_script.main(None)


MAX_MODELS_NUM = 3


def get_all_saved_models_timestamps(saved_models_dir):
    model_path, dir_base_name = (
        os.path.split(saved_models_dir)
        if os.path.split(saved_models_dir)[1] != ""
        else os.path.split(os.path.split(saved_models_dir)[0])
    )

    all_models = []
    for i in os.listdir(model_path):
        if dir_base_name in i:
            all_models.append(i)

    time_stamps = ["".join(i.split(dir_base_name + "-")) for i in all_models]

    datetime_objects = []
    for i in time_stamps:
        try:
            datetime_objects.append(mammut.str_parse_datetime(i))
        except Exception as ex:
            print(ex)

    return datetime_objects


def get_time_stamp_identifier(saved_model_dir, should_create_new):
    if should_create_new:
        # crea nuevo, obtiene todos elimina el mas viejo
        identifier = "-" + mammut.get_str_format_timezone_aware_datetime_now()
        all_timestamps = get_all_saved_models_timestamps(saved_model_dir)
        if len(all_timestamps) >= MAX_MODELS_NUM:
            older_model_identifier = "-" + mammut.str_format_datetime(
                min(all_timestamps)
            )
            model_path, dir_base_name = (
                os.path.split(saved_model_dir)
                if os.path.split(saved_model_dir)[1] != ""
                else os.path.split(os.path.split(saved_model_dir)[0])
            )
            older_model_name = model_path + "/" + dir_base_name + older_model_identifier
            print("delete model:", older_model_name)
            shutil.rmtree(older_model_name)
    else:
        # obtiene todos y retorna el mas nuevo
        all_timestamps = get_all_saved_models_timestamps(saved_model_dir)
        identifier = "-" + mammut.str_format_datetime(max(all_timestamps))

    return identifier


def load_labels(filename):
    with open(filename, "r") as f:
        return [line.strip() for line in f.readlines()]


def get_top_labels(
    tflite_output_file: str,
    image_file: str,
    label_file: str,
    top_count: int = 5,
    input_mean: float = 127.5,
    input_std: float = 127.5,
):
    interpreter = tf.lite.Interpreter(model_path=tflite_output_file)
    interpreter.allocate_tensors()

    input_details = interpreter.get_input_details()
    output_details = interpreter.get_output_details()

    # check the type of the input tensor
    floating_model = input_details[0]["dtype"] == np.float32

    # NxHxWxC, H:1, W:2
    height = input_details[0]["shape"][1]
    width = input_details[0]["shape"][2]
    img = Image.open(image_file).resize((width, height))

    # add N dim
    input_data = np.expand_dims(img, axis=0)

    if floating_model:
        input_data = (np.float32(input_data) - input_mean) / input_std

    interpreter.set_tensor(input_details[0]["index"], input_data)

    interpreter.invoke()

    output_data = interpreter.get_tensor(output_details[0]["index"])
    results = np.squeeze(output_data)

    labels = load_labels(label_file)
    if top_count > len(labels):
        top_count = len(labels)
    top_k = results.argsort()[-top_count:][::-1]
    top_labels = []
    for i in top_k:
        if floating_model:
            log.debug("{:08.6f}: {}".format(float(results[i]), labels[i]))
        else:
            log.debug("{:08.6f}: {}".format(float(results[i] / 255.0), labels[i]))
        top_labels.append((labels[i], results[i]))
    return top_labels


def label_image(
    tflite_output_file: str,
    image_file: str,
    label_file,
    top_count: int = 5,
    input_mean: float = 127.5,
    input_std: float = 127.5,
):
    top_labels = get_top_labels(
        tflite_output_file, image_file, label_file, top_count, input_mean, input_std
    )
    return top_labels[0][0]


def augmentate_and_retrain(directory: str, mammut_id: int, samples: int):
    base_dir_name = BASE_MODEL_VISON_DIR + "/" + str(mammut_id) + "/vision"
    augmentate_images(directory, samples)
    test_dir = directory
    training_sumaries_dir = base_dir_name + "/tf_summary"
    tflite_output_file = base_dir_name + "/saved_model.pb"
    labels_output_file = base_dir_name + "/labels_output_file.txt"
    retrianed_graph_folder = base_dir_name + "/models/place_recognition/"
    if not os.path.isdir(retrianed_graph_folder):
        os.makedirs(retrianed_graph_folder)
    retrain(
        test_dir,
        training_sumaries_dir,
        tflite_output_file,
        labels_output_file,
        retrianed_graph_folder,
    )
    return tflite_output_file, labels_output_file



from mammut.common.simulator_utilities import *
from ipywidgets import *


def get_operations_boxes(
    simulator_obj,
    users_rooms_spreadsheet_id,
    users_sheet_name,
    rooms_sheet_name,
    api_url,
    users_filename,
    rooms_filename,
    move_box_callback,
):

    # Unique fields in API for checking#
    unique_fields = [
        "facebook-id",
        "facebook-access-token",
        "main-phone-number",
        "main-email",
    ]

    def users_check_for_property_existence(prop_name: str, prop_val: str):
        user = simulator_obj.mammut_api.get_unique_vertex("user", prop_name, prop_val)
        if "error" in user:
            return {"error": "Error communicating with Mammut API"}

        if len(user) is 0:
            return {"user": False, "id": -1}
        else:
            return {"user": True, "id": user["id"]}

    # Select your mammut layout #
    mammut_user_dropdown = widgets.Dropdown(
        description="Mammut bot", disabled=False, button_style=""
    )
    mammut_user_dropdown.options = simulator_obj.users[UserType.machine.value]
    select_mammut_user_btn = widgets.Button(
        description="Seleccionar", button_style="primary"
    )

    def select_user_button_clicked(b):
        simulator_obj.current_user = simulator_obj.mammut_api.get_user(
            mammut_user_dropdown.value
        )
        move_box_callback()

    select_mammut_user_btn.on_click(select_user_button_clicked)
    select_mammut_box = widgets.VBox(
        [widgets.HBox([mammut_user_dropdown, select_mammut_user_btn])]
    )
    # /Select your mammut layout #

    # Create entities #
    bot_custom_properties_box = widgets.VBox(
        [
            widgets.HBox(
                [
                    widgets.Text(value="user-type", layout=Layout(width="50%")),
                    widgets.Text(value="machine", layout=Layout(width="50%")),
                ]
            )
        ]
    )
    add_bot_property_button = widgets.Button(
        description="+",
        button_style="primary",
        layout=Layout(width="35px", height="100%",),
        style={"button_color": "green"},
    )
    add_bot_output = widgets.HTML("")

    def add_bot_custom_property_button_clicked(b):
        new_property_name = widgets.Text(
            value="Property name", layout=Layout(width="50%")
        )
        new_property_value = widgets.Text(
            value="Property value", layout=Layout(width="50%")
        )
        bot_custom_properties_box.children = bot_custom_properties_box.children + tuple(
            [widgets.HBox([new_property_name, new_property_value])]
        )

    def create_bot_button_clicked(b):
        props_and_values = bot_custom_properties_box.children
        props_and_values_dic = {}
        for index in range(0, len(props_and_values)):
            props_and_values_dic[props_and_values[index].children[0].value] = (
                props_and_values[index].children[1].value
            )

        for i, j in props_and_values_dic.items():
            if i in unique_fields:
                exists = users_check_for_property_existence(i, j)
                print(exists)
                if exists["user"]:
                    bot_id = exists["id"]
                    add_user_output.value = (
                        f"User already exists with that {i}. User Id - {bot_id}"
                    )
                    return

        res = simulator_obj.mammut_api.post_request(
            simulator_obj.mammut_api.post_user_path, props_and_values_dic
        )
        if res.status_code == 200:
            res_json = res.json()
            created_user_id = -1
            for result in res_json["taskResult"]:
                if result["affectedElementName"] == "user":
                    created_user_id = result["affectedElementId"]
                    break
            add_bot_output.value = f"<h2>Bot created: ID-{created_user_id}</h2>"
            user = simulator_obj.mammut_api.get_user(created_user_id)
            add_new_user(simulator_obj, user, users_filename, UserType.machine.value)
        else:
            add_bot_output.value = "<h2>Error creating bot.</h2>"

    add_bot_property_button.on_click(add_bot_custom_property_button_clicked)
    create_bot_button = widgets.Button(
        description="Create bot",
        button_style="primary",
        layout=Layout(width="225px", height="100%"),
    )
    create_bot_button.on_click(create_bot_button_clicked)
    create_bot_box = widgets.VBox(
        [
            bot_custom_properties_box,
            add_bot_property_button,
            create_bot_button,
            add_bot_output,
        ],
        layout=Layout(width="100%", align_items="center"),
    )

    user_custom_properties_box = widgets.VBox(
        [
            widgets.HBox(
                [
                    widgets.Text(value="user-type", layout=Layout(width="50%")),
                    widgets.Text(value="normal", layout=Layout(width="50%")),
                ]
            )
        ]
    )
    add_user_property_button = widgets.Button(
        description="+",
        button_style="primary",
        layout=Layout(width="35px", height="100%",),
        style={"button_color": "green"},
    )
    add_user_output = widgets.HTML("")

    def add_user_custom_property_button_clicked(b):
        new_property_name = widgets.Text(
            value="Property name", layout=Layout(width="50%")
        )
        new_property_value = widgets.Text(
            value="Property value", layout=Layout(width="50%")
        )
        user_custom_properties_box.children = (
            user_custom_properties_box.children
            + tuple([widgets.HBox([new_property_name, new_property_value])])
        )

    def create_user_button_clicked(b):
        props_and_values = user_custom_properties_box.children
        props_and_values_dic = {}
        for index in range(0, len(props_and_values)):
            props_and_values_dic[props_and_values[index].children[0].value] = (
                props_and_values[index].children[1].value
            )

        for i, j in props_and_values_dic.items():
            if i in unique_fields:
                exists = users_check_for_property_existence(i, j)
                print(exists)
                if exists["user"]:
                    user_id = exists["id"]
                    add_user_output.value = (
                        f"User already exists with that {i}. User Id - {user_id}"
                    )
                    return

        if not "main-email" in props_and_values_dic:
            add_user_output.value = "<h2>Error: Es necesario el campo main-email.</h2>"
        else:
            res = simulator_obj.mammut_api.post_request(
                simulator_obj.mammut_api.post_user_path, props_and_values_dic
            )
            if res.status_code == 200:
                res_json = res.json()
                created_user_id = -1
                room_with_alex_id = -1
                for result in res_json["taskResult"]:
                    if result["affectedElementName"] == "user":
                        created_user_id = result["affectedElementId"]
                        break
                for result in res_json["taskResult"]:
                    if result["affectedElementName"] == "room":
                        room_with_alex_id = result["affectedElementId"]
                        break
                add_user_output.value = f"<h2>User created: ID-{created_user_id}</h2><h2>Room with Alex: ID-{room_with_alex_id}</h2>"
                user = simulator_obj.mammut_api.get_user(created_user_id)
                room_with_alex = simulator_obj.mammut_api.get_room(room_with_alex_id)
                add_new_user(
                    simulator_obj,
                    user,
                    users_filename,
                    props_and_values_dic["user-type"],
                )
                user_email = props_and_values_dic["main-email"]
                add_new_room(
                    simulator_obj,
                    room_with_alex,
                    rooms_filename,
                    "private",
                    f"_{user_email}",
                )
            else:
                add_user_output.value = "<h2>Error creating user.</h2>"

    add_user_property_button.on_click(add_user_custom_property_button_clicked)
    add_user_button = widgets.Button(
        description="Create user",
        button_style="primary",
        layout=Layout(width="225px", height="100%"),
    )
    add_user_button.on_click(create_user_button_clicked)
    create_user_box = widgets.VBox(
        [
            user_custom_properties_box,
            add_user_property_button,
            add_user_button,
            add_user_output,
        ],
        layout=Layout(width="100%", align_items="center"),
    )

    room_custom_properties_box = widgets.VBox(
        [
            widgets.HBox(
                [
                    widgets.Text(value="room-type", layout=Layout(width="50%")),
                    widgets.Text(value="group", layout=Layout(width="50%")),
                ]
            )
        ]
    )
    room_created_by_user_box = widgets.IntText(
        value=1234, description="Creator ID: ", style={"description_width": "initial"}
    )
    room_add_property_button = widgets.Button(
        description="+",
        button_style="primary",
        layout=Layout(width="35px", height="100%",),
        style={"button_color": "green"},
    )
    create_room_output = widgets.HTML("")

    def add_room_custom_property_button_clicked(b):
        new_property_name = widgets.Text(
            value="Property name", layout=Layout(width="50%")
        )
        new_property_value = widgets.Text(
            value="Property value", layout=Layout(width="50%")
        )
        room_custom_properties_box.children = (
            room_custom_properties_box.children
            + tuple([widgets.HBox([new_property_name, new_property_value])])
        )

    def create_room_button_clicked(b):
        props_and_values = room_custom_properties_box.children
        props_and_values_dict = {}
        for index in range(0, len(props_and_values)):
            if props_and_values[index].children[0].value == "has@user":
                props_and_values_dict[props_and_values[index].children[0].value] = {
                    "id": int(props_and_values[index].children[1].value)
                }
            else:
                props_and_values_dict[props_and_values[index].children[0].value] = (
                    props_and_values[index].children[1].value
                )
        create_room_path = simulator_obj.mammut_api.post_room_path.replace(
            "{userId}", str(room_created_by_user_box.value)
        )
        res = simulator_obj.mammut_api.post_request(
            create_room_path, props_and_values_dict
        )
        if res.status_code == 200:
            res_json = res.json()
            created_room_id = -1
            for result in res_json["taskResult"]:
                if result["affectedElementName"] == "room":
                    created_room_id = result["affectedElementId"]
                    break
            create_room_output.value = f"<h2>Room created: ID-{created_room_id}</h2>"
            room = simulator_obj.mammut_api.get_room(created_room_id)
            add_new_room(
                simulator_obj, room, rooms_filename, props_and_values_dict["room-type"]
            )
        else:
            create_room_output.value = "<h2>Error creating room.</h2>"

    room_add_property_button.on_click(add_room_custom_property_button_clicked)
    create_room_button = widgets.Button(
        description="Create Room",
        button_style="primary",
        layout=Layout(width="225px", height="100%"),
    )
    create_room_button.on_click(create_room_button_clicked)
    create_room_box = widgets.VBox(
        [
            room_custom_properties_box,
            room_add_property_button,
            room_created_by_user_box,
            create_room_button,
            create_room_output,
        ],
        layout=Layout(width="100%", align_items="center"),
    )

    create_entities_tab = widgets.Tab(
        children=[create_bot_box, create_user_box, create_room_box]
    )
    create_entities_tab.set_title(0, "Crear nuevo bot")
    create_entities_tab.set_title(1, "Crear nuevo user")
    create_entities_tab.set_title(2, "Crear nuevo room")
    # /Create entities #

    # load users & rooms#
    load_user_and_rooms_spreadsheet_id = widgets.Text(
        value=users_rooms_spreadsheet_id,
        description="Id Archivo:",
        layout=Layout(width="90%"),
        style={"description_width": "initial"},
    )
    users_sheetname = widgets.Text(
        value=users_sheet_name,
        description="Nombre del sheet de usuarios:",
        layout=Layout(width="80%", height="50px", pading="10px 15px 10px 8px"),
    )
    rooms_sheetname = widgets.Text(
        value=rooms_sheet_name,
        description="Nombre del sheet de chats:",
        layout=Layout(width="80%", height="50px", pading="10px 15px 10px 8px"),
    )
    load_users_btn = widgets.Button(
        description="Cargar Usuarios",
        button_style="primary",
        layout=Layout(width="225px", height="100%"),
    )
    load_rooms_btn = widgets.Button(
        description="Cargar Chats",
        button_style="primary",
        layout=Layout(width="225px", height="100%"),
    )
    users_output_html = widgets.HTML(value="")
    rooms_output_html = widgets.HTML(value="")

    def load_users_btn_clicked(b):
        mammut_api = Request(api_url)
        sheet_names = [s.strip() for s in users_sheet_name.value.split(",")]
        for sn in sheet_names:
            user_loader = UsersLoader(
                spreadsheet_id=load_user_and_rooms_spreadsheet_id.value,
                sheet_name=sn,
                mammut_api=mammut_api,
            )
            user_loader.load_users()

    def load_rooms_btn_clicked(b):
        mammut_api = Request(api_url)
        sheet_names = [s.strip() for s in rooms_sheet_name.value.split(",")]
        for sn in sheet_names:
            room_creator = RoomsCreator(
                simulator_obj.current_user,
                load_user_and_rooms_spreadsheet_id.value,
                mammut_api,
            )
            room_creator.activate(sn)

    load_users_btn.on_click(load_users_btn_clicked)
    load_rooms_btn.on_click(load_rooms_btn_clicked)

    load_users_rooms_box = widgets.VBox(
        [
            load_user_and_rooms_spreadsheet_id,
            widgets.HBox([users_sheetname, load_users_btn]),
            widgets.HBox([rooms_sheetname, load_rooms_btn]),
            users_output_html,
            rooms_output_html,
        ]
    )
    # /Cargar usuarios y rooms #

    # Cargar videos #
    video_url_textbox = widgets.Text(
        value="", description="Url:", layout=Layout(width="100%", height="50px")
    )
    video_name_textbox = widgets.Text(
        value="", description="Nombre:", layout=Layout(width="100%", height="50px")
    )
    save_video_btn = widgets.Button(
        description="Guardar video",
        button_style="primary",
        layout=Layout(width="225px", height="100%"),
    )

    def save_video_button_clicked(b):
        if video_name_textbox.value != "" and video_url_textbox.value != "":
            post_result = simulator_obj.mammut_api.post_video(
                video_name_textbox.value,
                video_url_textbox.value,
                simulator_obj.current_user.id,
            )
        else:
            print("Complete los campos")
        video_url_textbox.value = ""
        video_name_textbox.value = ""

    save_video_btn.on_click(save_video_button_clicked)
    register_video_box = widgets.VBox(
        [video_url_textbox, video_name_textbox, save_video_btn]
    )
    # /Cargar videos #

    # Cargar Calendario #
    style = {"description_width": "initial"}
    calendar_config_layout = Layout(
        width="80%", height="50px", pading="10px 15px 10px 8px"
    )

    calendar_title_textbox = widgets.Text(
        value="", description="Nombre:", layout=calendar_config_layout, style=style
    )

    calendar_supervisor_email_textbox = widgets.Text(
        value="",
        description="Supervisor User email:",
        layout=calendar_config_layout,
        style=style,
    )
    places_textboxes = widgets.VBox([])

    def add_places_textbox(index):
        posible_places_textbox = widgets.Text(
            value="",
            description="Lugar " + str(index + 1) + ":",
            layout=calendar_config_layout,
            style=style,
        )
        places_textboxes.children = list(places_textboxes.children) + [
            posible_places_textbox
        ]

    def add_place_button_clicked(b):
        add_places_textbox(len(places_textboxes.children))

    add_places_textbox(len(places_textboxes.children))
    add_place_button = widgets.Button(
        description="Agregar Lugar",
        button_style="primary",
        layout=Layout(width="225px", height="100%", margin="10px 15px 10px 8px"),
    )
    add_place_button.on_click(add_place_button_clicked)
    posibles_places_format_dict = dict()
    for i in PosiblePlacesFormat:
        posibles_places_format_dict[i.name] = i.value
    posibles_places_format = widgets.Dropdown(
        options=posibles_places_format_dict,
        value=0,
        layout=calendar_config_layout,
        description="Formato Campo Lugares:",
        style=style,
    )

    country_code_dict = dict()
    for i in CountryCodes:
        country_code_dict[i.name + "(" + i.value + ")"] = i.value
    country_code_selector = widgets.Dropdown(
        options=country_code_dict,
        value=CountryCodes.US.value,
        description="Coding de Pais:",
        layout=calendar_config_layout,
        style=style,
    )

    languages_options = dict()
    for i in RegionalSettingsType:
        languages_options[i.name] = i.value

    languages_box = widgets.VBox([])

    def add_language_to_box(index):
        language_dropdown = widgets.Dropdown(
            options=languages_options,
            value="es",
            description="Idioma " + str(index + 1) + ":",
            layout=calendar_config_layout,
            style=style,
        )
        languages_box.children = list(languages_box.children) + [language_dropdown]

    add_language_to_box(len(languages_box.children))
    add_language_button = widgets.Button(
        description="Agregar Idioma",
        button_style="primary",
        layout=Layout(width="225px", height="100%", margin="10px 15px 10px 8px"),
    )

    def add_language_button_clicked(b):
        add_language_to_box(len(languages_box.children))

    add_language_button.on_click(add_language_button_clicked)

    ## CUSTOM FIELDS SUBSECTION ##
    custom_fields_box = widgets.VBox([])

    def add_custom_fields_section(index):
        custom_fields_section_title = widgets.HTML(
            value="<b>Campos Adicionales " + str(index + 1) + "</b>"
        )
        custom_data_name = widgets.Text(
            value="",
            description="Nombre del Campo:",
            layout=calendar_config_layout,
            style=style,
        )

        custom_data_values_textboxes = widgets.VBox([])

        def add_custom_data_values_textboxes(index):
            custom_data_value = widgets.Text(
                value="",
                description="Valor del Campo " + str(index + 1) + ":",
                layout=calendar_config_layout,
                style=style,
            )
            custom_data_values_textboxes.children = list(
                custom_data_values_textboxes.children
            ) + [custom_data_value]

        add_custom_data_values_textboxes(len(custom_data_values_textboxes.children))
        add_custom_data_values_button = widgets.Button(
            description="Agregar Valor",
            button_style="primary",
            layout=Layout(width="225px", height="100%", margin="10px 15px 10px 8px"),
        )

        def add_custom_data_values_button_clicked(b):
            add_custom_data_values_textboxes(len(custom_data_values_textboxes.children))

        add_custom_data_values_button.on_click(add_custom_data_values_button_clicked)

        def add_videos_selector():
            response = simulator_obj.mammut_api.get_videos(
                simulator_obj.current_user.id
            )
            videos_dict = dict()
            print(response)
            for i in response:
                videos_dict[i["title"]] = i["id"]
            videos_dropdown = widgets.Dropdown(
                options=videos_dict,
                value=response[0]["id"],
                description="Videos Disponibles:",
                layout=calendar_config_layout,
                style=style,
            )
            if (
                len(custom_data_template_box.children) <= 6
            ):  # TODO:como soluciono esto?? parece que se activa dos veces el callback
                custom_data_template_box.children = list(
                    custom_data_template_box.children
                ) + [videos_dropdown]

        custom_data_format = widgets.Dropdown(
            options=posibles_places_format_dict,
            value=0,
            description="Formato del Campo:",
            layout=calendar_config_layout,
            style=style,
        )

        def custom_data_format_change(b):
            if b["new"] == PosiblePlacesFormat.video.value:
                add_videos_selector()

        custom_data_format.observe(custom_data_format_change)

        mandatory_widget = widgets.Checkbox(
            value=False, description="Requerido", disabled=False
        )

        custom_data_template_box = widgets.VBox(
            [
                custom_fields_section_title,
                custom_data_name,
                custom_data_values_textboxes,
                add_custom_data_values_button,
                custom_data_format,
                mandatory_widget,
            ],
            layout=Layout(
                display="flex",
                flex_flow="column",
                border="solid 1px",
                align_items="stretch",
                width="100%",
                pading="10px 15px 10px 8px",
            ),
        )
        custom_fields_box.children = list(custom_fields_box.children) + [
            custom_data_template_box
        ]

    add_custom_fields_section(len(custom_fields_box.children))
    add_custom_section_button = widgets.Button(
        description="Agregar seccion de Campos Adicionales",
        button_style="primary",
        layout=Layout(width="225px", height="100%", margin="10px 15px 10px 8px"),
    )

    def add_custom_section_button_clicked(b):
        add_custom_fields_section(len(custom_fields_box.children))

    add_custom_section_button.on_click(add_custom_section_button_clicked)

    available_reminder_varibales = widgets.HTML(
        value="&lt;date&gt;&lt;arrive&gt;&lt;place&gt;&lt;language&gt;",
        description="Variables disponibles: ",
        style={"description_width": "initial"},
    )
    reminder_text_templates_box = widgets.VBox(
        [available_reminder_varibales],
        layout=Layout(
            display="flex",
            flex_flow="column",
            border="solid 1px",
            align_items="stretch",
            width="100%",
            pading="10px 15px 10px 8px",
        ),
    )

    def add_reminder_text_template(index=0):
        reminder_text_area = widgets.Textarea(
            value="",
            placeholder="Mensaje Recordatorio",
            description="Texto Recordatorio " + str(index + 1) + ":",
            disabled=False,
            layout=calendar_config_layout,
        )
        reminder_text_lang_dropdown = widgets.Dropdown(
            options=languages_options,
            value="es",
            description="Idioma Recordatorio:",
            layout=calendar_config_layout,
            style=style,
        )
        reminder_text_templates_elem = widgets.VBox(
            [reminder_text_area, reminder_text_lang_dropdown]
        )
        reminder_text_templates_box.children = list(
            reminder_text_templates_box.children
        ) + [reminder_text_templates_elem]

    add_reminder_text_template(len(reminder_text_templates_box.children))
    add_reminder_button = widgets.Button(
        description="Agrgar Recordatorio",
        button_style="primary",
        layout=Layout(width="225px", height="100%", margin="10px 15px 10px 8px"),
    )

    def add_reminder_button_clicked(b):
        add_reminder_text_template(len(reminder_text_templates_box.children))

    add_reminder_button.on_click(add_reminder_button_clicked)

    save_calendar_button = widgets.Button(
        description="Guardar Calendario",
        button_style="primary",
        layout=Layout(width="225px", height="100%", margin="10px 15px 10px 8px"),
    )

    def save_calendar_button_clicked(b):
        #  calendar = Calendar(custom_data_name.value)
        title = calendar_title_textbox.value
        supervisor_user_by_mail = simulator_obj.mammut_api.get_user_by_mail(
            calendar_supervisor_email_textbox.value
        )
        if supervisor_user_by_mail is not None:
            admin_user_id_value = supervisor_user_by_mail.id
        else:
            print(
                "error, no hay supervisor para el email",
                calendar_supervisor_email_textbox.value,
            )
        posibles_places_values = [i.value for i in places_textboxes.children]
        posibles_places_format_value = posibles_places_format.value
        country_code_selector_value = country_code_selector.value
        regional_setings_values = [i.value for i in languages_box.children]
        custom_sections_values = dict()
        children_list = [i for i in custom_fields_box.children]
        elems = [i for i in children_list]
        a = []
        for i in children_list:
            a.append(i)
        for custom_section in list(children_list):
            custom_data_name_value = custom_section.children[1].value
            custom_data_values_textboxes_values = [
                i.value for i in custom_section.children[2].children
            ]
            custom_data_format_value = custom_section.children[4].value
            if custom_data_format_value == PosiblePlacesFormat.video.value:
                custom_data_values_textboxes_values[0] = (
                    custom_data_values_textboxes_values[0]
                    + "/video?id="
                    + str(custom_section.children[6].value)
                )
            mandatory_widget_value = custom_section.children[5].value
            custom_fields_values = CalendarCustomValues(
                name=custom_data_name_value,
                value=custom_data_values_textboxes_values,
                value_type=custom_data_format_value,
                mandatory=mandatory_widget_value,
            )
            custom_sections_values[
                custom_data_name_value
            ] = custom_fields_values.to_json()[custom_data_name_value]
        reminders = {}
        for i, vbox in enumerate(list(reminder_text_templates_box.children)[1:]):
            reminders[vbox.children[1].value] = vbox.children[0].value
        calendar = Calendar(
            user_id=admin_user_id_value,
            mammut_user=simulator_obj.current_user,
            title=title,
            possible_places=posibles_places_values,
            possible_places_format=posibles_places_format_value,
            possible_regional_settings=regional_setings_values,
            custom_fields=custom_sections_values,
            reminder_text_templates=reminders,
            country_code=country_code_selector_value,
        )

        response = simulator_obj.mammut_api.post_calendar(calendar)
        print("response", response)
        if response["status"] != "Failed":
            print("reponse\n", response)
        else:
            print("Error creating calendar")

    save_calendar_button.on_click(save_calendar_button_clicked)
    register_calendar_event_content_box = widgets.VBox(
        [
            calendar_title_textbox,
            calendar_supervisor_email_textbox,
            places_textboxes,
            add_place_button,
            posibles_places_format,
            country_code_selector,
            languages_box,
            add_language_button,
            custom_fields_box,
            add_custom_section_button,
            reminder_text_templates_box,
            add_reminder_button,
            save_calendar_button,
        ]
    )
    # /Cargar Calendario #

    return tuple(
        [
            select_mammut_box,
            create_entities_tab,
            load_users_rooms_box,
            register_video_box,
            register_calendar_event_content_box,
        ]
    )



# coding=utf-8

import logging
from mammut.common.storage.storage_base import SlidesStorageBase
from pptx import Presentation
from pptx.enum.shapes import MSO_SHAPE_TYPE
from collections import namedtuple

logger = logging.getLogger()

SlideElement = namedtuple("SlideElement", ["slideNumber", "elementType", "content"])


class EmbeddedStoragePowerPoint(SlidesStorageBase):
    slides_metadata_re_string = (
        "^(sheet-id|Sheet-id|SHEET-ID)[ ]*->[ ]*(.+\\.(xlsx|pptx))$"
    )

    def __init__(self, storage_path):
        SlidesStorageBase.__init__(self, storage_path)

    def get_slide_elements(self, presentation_id):
        prs = Presentation(self.storage_path + presentation_id)
        slide_elements = []
        for slide in prs.slides:
            for j, shape in enumerate(slide.shapes):
                if shape.shape_type == MSO_SHAPE_TYPE.PICTURE:
                    url = shape.click_action.hyperlink.address
                    slide_elements.append((SlideElement(j + 1, "image", url)))
                elif shape.has_text_frame:
                    text = ""
                    for paragraph in shape.text_frame.paragraphs:
                        for run in paragraph.runs:
                            text = text + run.text

                    slide_elements.append((SlideElement(j + 1, "text", text),))

        return slide_elements



from mammut.common.util import StringEnum
from mammut.common.lexicon.paradigm_producer import ParadigmProducerGroup


class POSType(StringEnum):
    Entity = "entity"
    Event = "event"


class POSdescriptor:
    POS_CLASS_GROUP_COLUMN_NAME = "pos-class-group"
    POS_CLASS_COLUMN_NAME = "pos-class"
    POS_NAME_COLUMN_NAME = "pos-name"
    POS_ID_COLUMN_NAME = "pos-id"
    POS_TYPE_COLUMN_NAME = "pos-type"
    MORPHEME_PARADIGMS_COLUMN_NAME = "morph-paradigms"
    MISSPELLING_PARADIGMS_COLUMN_NAME = "miss-paradigms"
    FEATURES_COLUMN_NAME = "features"
    REGIONAL_SETTINGS_COLUMN_NAME = "regional-settings"
    ARC_COLUMN_NAME = "arc-name"
    POS_ABBREVIATION = "pos-abbreviation"
    POS_SHIFTS = "pos-shifts"

    SHEETS_LOADED = {}

    def __init__(
        self,
        class_group: str,
        pos_class: str,
        name: str,
        id: str,
        pos_type: str,
        morpheme_paradigm_sheet_titles: str,
        misspelling_paradigm_sheet_titles: str,
        feature_paradigm_sheet_titles: str,
        regional_settings: str,
        arc_name: str,
        pos_abbreviation: str,
        pos_shifts: str,
        standard,
    ):
        self.class_group = class_group
        self.pos_class = pos_class
        self.name = name
        self.id = id
        self.pos_type = pos_type
        self.used_regional_settings_list = []
        self.morpheme_paradigms = {}
        self.misspelling_paradigms = {}
        self.feature_paradigms = {}
        self.regional_settings_list = [
            s.strip() for s in regional_settings.split(",") if s.strip()
        ]

        if POSType.contains_value(pos_type):
            self.pos_type = POSType(pos_type)

        self.morpheme_paradigm_sheet_titles_list = [
            s.strip() for s in morpheme_paradigm_sheet_titles.split(",") if s.strip()
        ]
        self.misspelling_paradigm_sheet_titles_list = [
            s.strip() for s in misspelling_paradigm_sheet_titles.split(",") if s.strip()
        ]
        self.feature_paradigm_sheet_titles_list = [
            s.strip() for s in feature_paradigm_sheet_titles.split(",") if s.strip()
        ]
        self.arc_name = arc_name
        self.pos_abbreviation = pos_abbreviation
        self.pos_shifts_ids = [s.strip() for s in pos_shifts.split(",") if s.strip()]
        self.pos_shifts = []
        self.standard = standard
        for (
            morpheme_paradigm_sheet_title_and_rs
        ) in self.morpheme_paradigm_sheet_titles_list:
            if (
                morpheme_paradigm_sheet_title_and_rs != ""
                and ":" in morpheme_paradigm_sheet_title_and_rs
            ):
                regional_setting = morpheme_paradigm_sheet_title_and_rs.split(":")[0]
                morpheme_paradigm_sheet_title = morpheme_paradigm_sheet_title_and_rs.split(
                    ":"
                )[
                    1
                ]
                self.used_regional_settings_list.append(regional_setting)
            elif morpheme_paradigm_sheet_title_and_rs != "":
                if self.regional_settings_list.__len__() == 0:
                    regional_setting = ""
                    morpheme_paradigm_sheet_title = morpheme_paradigm_sheet_title_and_rs
                else:
                    for setting in self.regional_settings_list:
                        if setting in self.used_regional_settings_list:
                            continue
                        if (
                            morpheme_paradigm_sheet_title_and_rs
                            not in POSdescriptor.SHEETS_LOADED
                        ):
                            self.morpheme_paradigms[setting] = ParadigmProducerGroup(
                                morpheme_paradigm_sheet_title_and_rs,
                                self.standard,
                                setting,
                            )
                            POSdescriptor.SHEETS_LOADED[
                                morpheme_paradigm_sheet_title_and_rs
                            ] = self.morpheme_paradigms[setting]
                        else:
                            self.morpheme_paradigms[
                                setting
                            ] = POSdescriptor.SHEETS_LOADED[
                                morpheme_paradigm_sheet_title_and_rs
                            ]
                    continue
            else:
                regional_setting = ""
                morpheme_paradigm_sheet_title = ""
            if morpheme_paradigm_sheet_title not in POSdescriptor.SHEETS_LOADED:
                self.morpheme_paradigms[regional_setting] = ParadigmProducerGroup(
                    morpheme_paradigm_sheet_title, self.standard, regional_setting
                )
                POSdescriptor.SHEETS_LOADED[
                    morpheme_paradigm_sheet_title
                ] = self.morpheme_paradigms[regional_setting]
            else:
                self.morpheme_paradigms[regional_setting] = POSdescriptor.SHEETS_LOADED[
                    morpheme_paradigm_sheet_title
                ]
        self.used_regional_settings_list = []
        for (
            misspelling_paradigm_sheet_title_and_rs
        ) in self.misspelling_paradigm_sheet_titles_list:
            if (
                misspelling_paradigm_sheet_title_and_rs != ""
                and ":" in misspelling_paradigm_sheet_title_and_rs
            ):
                regional_setting = misspelling_paradigm_sheet_title_and_rs.split(":")[0]
                misspelling_paradigm_sheet_title = misspelling_paradigm_sheet_title_and_rs.split(
                    ":"
                )[
                    1
                ]
                self.used_regional_settings_list.append(regional_setting)
            elif misspelling_paradigm_sheet_title_and_rs != "":
                if self.regional_settings_list.__len__() == 0:
                    regional_setting = ""
                    misspelling_paradigm_sheet_title = (
                        misspelling_paradigm_sheet_title_and_rs
                    )
                else:
                    for setting in self.regional_settings_list:
                        if setting in self.used_regional_settings_list:
                            continue
                        if (
                            misspelling_paradigm_sheet_title_and_rs
                            not in POSdescriptor.SHEETS_LOADED
                        ):
                            self.misspelling_paradigms[setting] = ParadigmProducerGroup(
                                misspelling_paradigm_sheet_title_and_rs,
                                self.standard,
                                setting,
                            )
                            POSdescriptor.SHEETS_LOADED[
                                misspelling_paradigm_sheet_title_and_rs
                            ] = self.misspelling_paradigms[setting]
                        else:
                            self.misspelling_paradigms[
                                setting
                            ] = POSdescriptor.SHEETS_LOADED[
                                misspelling_paradigm_sheet_title_and_rs
                            ]
                    continue
            else:
                regional_setting = ""
                misspelling_paradigm_sheet_title = ""
            if misspelling_paradigm_sheet_title not in POSdescriptor.SHEETS_LOADED:
                self.misspelling_paradigms[regional_setting] = ParadigmProducerGroup(
                    misspelling_paradigm_sheet_title, self.standard, regional_setting
                )
                POSdescriptor.SHEETS_LOADED[
                    misspelling_paradigm_sheet_title
                ] = self.misspelling_paradigms[regional_setting]
            else:
                self.misspelling_paradigms[
                    regional_setting
                ] = POSdescriptor.SHEETS_LOADED[misspelling_paradigm_sheet_title]
        self.used_regional_settings_list = []
        for (
            feature_paradigm_sheet_title_and_rs
        ) in self.misspelling_paradigm_sheet_titles_list:
            if (
                feature_paradigm_sheet_title_and_rs != ""
                and ":" in feature_paradigm_sheet_title_and_rs
            ):
                regional_setting = feature_paradigm_sheet_title_and_rs.split(":")[0]
                feature_paradigm_sheet_title = feature_paradigm_sheet_title_and_rs.split(
                    ":"
                )[
                    1
                ]
                self.used_regional_settings_list.append(regional_setting)
            elif feature_paradigm_sheet_title_and_rs != "":
                if self.regional_settings_list.__len__() == 0:
                    regional_setting = ""
                    feature_paradigm_sheet_title = feature_paradigm_sheet_title_and_rs
                else:
                    for setting in self.regional_settings_list:
                        if setting in self.used_regional_settings_list:
                            continue
                        if (
                            feature_paradigm_sheet_title_and_rs
                            not in POSdescriptor.SHEETS_LOADED
                        ):
                            self.feature_paradigms[setting] = ParadigmProducerGroup(
                                feature_paradigm_sheet_title_and_rs,
                                self.standard,
                                setting,
                            )
                            POSdescriptor.SHEETS_LOADED[
                                feature_paradigm_sheet_title_and_rs
                            ] = self.feature_paradigms[setting]
                        else:
                            self.feature_paradigms[
                                setting
                            ] = POSdescriptor.SHEETS_LOADED[
                                feature_paradigm_sheet_title_and_rs
                            ]
                    continue
            else:
                regional_setting = ""
                feature_paradigm_sheet_title = ""
            if feature_paradigm_sheet_title not in POSdescriptor.SHEETS_LOADED:
                self.feature_paradigms[regional_setting] = ParadigmProducerGroup(
                    feature_paradigm_sheet_title, self.standard, regional_setting
                )
                POSdescriptor.SHEETS_LOADED[
                    feature_paradigm_sheet_title
                ] = self.feature_paradigms[regional_setting]
            else:
                self.feature_paradigms[regional_setting] = POSdescriptor.SHEETS_LOADED[
                    feature_paradigm_sheet_title
                ]

    def set_pos_references(self, pos_by_id: dict):
        for pos_id in self.pos_shifts_ids:
            self.pos_shifts.append(pos_by_id[pos_id])

    def __str__(self):
        return f"{self.name}"



import time
import paramiko

class SshCommands:
    def __init__(self, host, port, user, key_filename):
        self.host = host
        self.port = port
        self.user = user
        self.key_filename = key_filename

    def get_connection(self):
        ssh = paramiko.SSHClient()  # Iniciamos un cliente SSH
        ssh.load_system_host_keys() #cargamos hosts conocidos
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(self.host, port=self.port, username=self.user, key_filename=self.key_filename) #generamos conexion
        shell = ssh.invoke_shell() #creamos canal seguro
        return shell 

    def send_string_and_wait(self, command, wait_time, should_print, shell):
        # Send the command
        shell.send(command)
        # Wait a bit, if necessary
        while not shell.recv_ready():
            time.sleep(wait_time)
        # Flush the receive buffer
        receive_buffer = shell.recv(4096)
        # Print the receive buffer, if necessary
        if should_print:
            print(str(receive_buffer,'utf-8'))

    def restart_service(self, service, password):
        shell = self.get_connection()
        #update service
        self.send_string_and_wait("sudo docker service update --env-add UPDATE=1 " + service + "\n", 2, False, shell)
        #su passwd
        self.send_string_and_wait(password + "\n", 2, True, shell)
        shell.close()



# coding=utf-8
from typing import List

from sklearn import neighbors
import tensorflow as tf
import tensorflow_hub as hub
import tf_sentencepiece  # Not used directly but needed to import TF ops.
from simpleneighbors import SimpleNeighbors
from tensorflow.python import global_variables_initializer, tables_initializer, Session
from tensorflow.python.keras.backend import placeholder

from mammut.common.package import Package

import logging

log = logging.getLogger(__name__)


class Paraphrases:
    NUM_RESULTS = 10

    def __init__(self, package: Package):
        self.package = package
        module_url = (
            "https://tfhub.dev/google/universal-sentence-encoder-multilingual/1"
        )
        # Set up graph.
        g = tf.Graph()
        with g.as_default():
            self.text_input = placeholder(dtype=tf.string, shape=[None])
            multiling_embed = hub.Module(module_url)
            self.embedded_text = multiling_embed(self.text_input)
            init_op = tf.group([global_variables_initializer(), tables_initializer()])
        g.finalize()
        # Initialize session.
        self.session = Session(graph=g)
        self.session.run(init_op)
        self.indexes = {}

    def get_similar_phrases(
        self, corpus_name: str, corpus_type: str, query: str
    ) -> List[str]:
        """Look in a corpus for all phrases similar to one provided.

        :param corpus_name: Name of the corpus to check.
        :param corpus_type: Type of the corpus to check.
        :param query: Phrase which we are looking similarities for.
        :return: List of Phrases found in the corpus specified.
        """
        index_by_corpus_name = {}
        if corpus_type in self.indexes:
            index_by_corpus_name = self.indexes[corpus_type]
        else:
            self.indexes[corpus_type] = index_by_corpus_name
        index = None
        if corpus_name in index_by_corpus_name:
            index = index_by_corpus_name[corpus_name]
        else:
            corpus = self.package.corpus_map.get_corpus(corpus_type, corpus_name)
            all_events = [
                tm.original_text
                for s in corpus.sceneries.values()
                for e in s.events
                for tm in e.tokenized_messages
            ]
            events_embeddings = self.session.run(
                self.embedded_text, feed_dict={self.text_input: all_events}
            )
            num_index_trees = 40
            embedding_dimensions = len(events_embeddings[0])
            metric = None
            metrics = ["dot", "cosine", "euclidean"]
            for m in metrics:
                if m in neighbors.VALID_METRICS["brute"]:
                    metric = m
                    break
            if not metric:
                raise RuntimeError("None of the valid metrics were found in sklearn.")
            index = SimpleNeighbors(embedding_dimensions, metric=metric)
            for i in range(len(all_events)):
                index.add_one(all_events[i], events_embeddings[i])
            index.build(n=num_index_trees)
        query_embedding = self.session.run(
            self.embedded_text, feed_dict={self.text_input: [query]}
        )[0]
        search_results = index.nearest(query_embedding, n=Paraphrases.NUM_RESULTS)
        return search_results



# coding=utf-8
"""
This file contains implementations related to the Gumbel Distribution

1.- Continuos aproximations to the Gumbel Max Trick: the implementation was taken from the blog post
 http://blog.evjang.com/2016/11/tutorial-categorical-variational.html. Which is a short presentations of the
 main results of the paper https://arxiv.org/pdf/1611.01144.pdf
 A introductory explanation of the trick can be found in
 -  https://laurent-dinh.github.io/2016/11/22/gumbel-max.html
 -  https://hips.seas.harvard.edu/blog/2013/04/06/the-gumbel-max-trick-for-discrete-distributions/
"""

import tensorflow as tf


def sample_gumbel(shape, eps=1e-20):
    """Sample from Gumbel(0, 1)"""
    #TODO: Use memoization to avoid calls to log
    u = tf.random_uniform(shape, minval=0, maxval=1)
    return -tf.log(-tf.log(u + eps) + eps)


def gumbel_softmax_sample(logits, temperature):
    """ Draw a sample from the Gumbel-Softmax distribution"""
    y = logits + sample_gumbel(tf.shape(logits))
    return tf.nn.softmax(y / temperature)


def gumbel_softmax(logits, temperature, hard=False):
    """Sample from the Gumbel-Softmax distribution and optionally discretize.
    Args:
      options_probabilities: [batch_size, probability] unnormalized log-probs
      temperature: non-negative scalar
      hard: if True, take argmax, but differentiate w.r.t. soft sample y
    Returns:
      [batch_size, n_class] sample from the Gumbel-Softmax distribution.
      If hard=True, then the returned sample will be one-hot, otherwise it will
      be a probabilitiy distribution that sums to 1 across classes
    """
    y = gumbel_softmax_sample(logits, temperature)
    if hard:
        # k = tf.shape(logits)[-1]
        # y_hard = tf.cast(tf.one_hot(tf.argmax(y,1),k), y.dtype)
        y_hard = tf.cast(tf.equal(y, tf.reduce_max(y, 1, keepdims=True)), y.dtype)
        y = tf.stop_gradient(y_hard - y) + y
    return y


def gumbel_softmax_option(logits, temperature, options):
    """Sample from the Gumbel-Softmax distribution and get the option index.
    Args:
      options_probabilities: [batch_size, probability] unnormalized log-probs
      temperature: non-negative scalar
    Returns:
      [batch_size, n_class_id] sample from the Gumbel-Softmax distribution.
    """
    y = gumbel_softmax(logits, temperature, hard=True)
    y = tf.reduce_sum(options * y, 1)
    return y


from mammut.common.basic_tokenizer import BasicTokenizer
from mammut.common.corpus.constants import CorpusTable
from mammut.common.corpus.corpus_base import Corpus, CorpusScenarioEntry
from mammut.common.lexicon.dictionary import Dictionary


class NaturalCorpus(Corpus):
    def __init__(
        self,
        spreadsheet_id: str,
        id,
        corpus_sheet_title: str,
        corpus_extension_sheet_title: str,
        tokenizer: BasicTokenizer,
        dictionary: Dictionary,
        regional_settings: str,
        annotable: bool,
        corpus_dataframe=None,
    ):
        Corpus.__init__(
            self,
            spreadsheet_id,
            id,
            corpus_sheet_title,
            corpus_extension_sheet_title,
            tokenizer,
            dictionary,
            regional_settings,
            annotable,
            corpus_dataframe=corpus_dataframe,
        )

    def get_corpus_scenario_entry_from_row(self, g):
        return CorpusScenarioEntry(
            g[0],
            g[1].iloc[0][CorpusTable.SCENARIO_TYPE_COLUMN_NAME()],
            g[1],
            self.tokenizer,
            self.dictionary,
            self.regional_settings,
            self,
        )



# coding=utf-8
from typing import List, Optional, Set
from mammut.curriculum.core.curriculum_base import CurriculumBase
from mammut.curriculum.classroom.single_classroom import SingleClassroom
from mammut.curriculum.course.course import Course
from mammut.curriculum.core.curriculum_state_machine import (
    CurriculumStateMachineMixin,
)
import mammut.curriculum.report as report


class Curriculum(CurriculumBase, CurriculumStateMachineMixin):
    """Models the Curriculum learning functionality in a single classroom
    environment.

    A Curriculum is composed of a set of courses.
    Each course attempts to train language competences that belongs
    to a language level (https://doc.mammut.io/docs_es/framework/language_arquitecture/).

    The courses in a Curriculum forms a flowchart, which is
    a directed acyclic graph. The flowchart implies that courses are
    meant to be progressive, which allows to organize them into academic
    periods.

    Courses belonging to the same academic period, can be processed
    concurrently.

    This single curriculum doesn't support KtT interpretation.
    """

    def __init__(
        self,
        package_id: str,
        standard_spreadsheet_id: str,
        curriculum_base_sheet_title: str = "curriculum",
        lessons_base_sheet_title: str = "lessons",
        assessments_sheet_title: str = "assessments",
        standard_base_sheet_title: str = "base",
        standard_dict_sources_sheet_title: str = "dict-sources",
        standard_pos_sheet_title: str = "pos-tags",
        standard_features_sheet_title: str = "features",
        standard_dimensions_sheet_title: str = "dimensions",
        dictionary_sheet_title: str = "dictionary",
        standard_functionals_sheet_title: str = "functionals",
        standard_non_functionals_sheet_title: str = "non-functionals",
        standard_primes_sheet_title: str = "primes",
        standard_synonym_sheet_title: str = "synonym",
        synonym_sheet_title: str = "synonym",
        progress_sheet_title: str = "progress",
        package_id_is_spreadsheet: bool = False,
        show_progress: bool = True,
    ):
        super(Curriculum, self).__init__(
            package_id,
            standard_spreadsheet_id,
            curriculum_base_sheet_title,
            lessons_base_sheet_title,
            assessments_sheet_title,
            standard_base_sheet_title,
            standard_dict_sources_sheet_title,
            standard_pos_sheet_title,
            standard_features_sheet_title,
            standard_dimensions_sheet_title,
            dictionary_sheet_title,
            standard_functionals_sheet_title,
            standard_non_functionals_sheet_title,
            standard_primes_sheet_title,
            standard_synonym_sheet_title,
            synonym_sheet_title,
            progress_sheet_title,
            package_id_is_spreadsheet,
            show_progress,
            enable_ktt_interpretation=False,
        )
        self._next_academic_period_courses: Optional[Set[Course]] = None
        self.classroom: SingleClassroom = None
        """
            Attribute flag to indicate auto transition
            after first academic period has been executed.
        """
        self._started: bool = False

    def _compute_next_academic_period(self):
        """Select the set of courses for the next academic period.
        """
        courses = list(self.flowchart.graph.nodes)
        approved_courses: List[Course] = []
        unapproved_courses: List[Course] = []

        """
            Courses that are already approved don't count. 
        """
        for c in courses:
            if not c.approved():
                unapproved_courses.append(c)
            else:
                approved_courses.append(c)

        """
            All courses approved.
        """
        if not unapproved_courses:
            report.send_message(
                report.CurriculumGeneralInfoMessage(
                    "All courses has been approved. Curriculum Completed."
                )
            )
            self._next_academic_period_courses = []
            self.trigger(self.CURRICULUM_COMPLETED_TRIGGER)
            return

        """
            Computes the next academic period based on 
            the minimum academic period course(s) that haven't be approved. 
            Only courses with the same academic period can be concurrently processed.
        """
        min_academic_period_course = min(
            unapproved_courses, key=lambda course: course.academic_period
        )

        """
            Assign the unapproved courses from the lowest academic period 
        """
        self._next_academic_period_courses = [
            c
            for c in unapproved_courses
            if c.academic_period == min_academic_period_course.academic_period
        ]

        """
            At this point, it is possible to decide if the curriculum 
            failed. If at least one course from the next academic period
            is not allowed, the curriculum failed. 
        """
        if any(
            list(
                map(
                    lambda course: not course.is_allowed(),
                    self._next_academic_period_courses,
                )
            )
        ):
            report.send_message(
                report.CurriculumGeneralWarningMessage(
                    "Curriculum process failed. Maximum retries reached for course."
                )
            )
            self.trigger(self.FAIL_COURSE_TRIGGER)
            return

        """
            At this point, it's safe to make the transition to the 
            next academic period processing cycle. 
        """
        report.send_message(
            report.CurriculumGeneralInfoMessage(
                f"New academic period. [{min_academic_period_course.academic_period}] academic period "
                f"with [{len(self._next_academic_period_courses)}] assigned courses"
            ),
        )
        if self._started:
            self.trigger(self.BEGIN_COURSES_TRIGGER)

    def _all_courses_completed(self) -> bool:
        """Returns False if there's at least one unapproved.
        """
        return len(self._next_academic_period_courses) == 0

    def _take_courses(self):
        """This is a wrapper function for course executor
        take_courses method.

        It is configured to be invoked when the curriculum
        makes the transition to ASSIGNED_STATE.
        """
        self.classroom.take_courses(
            set(
                map(lambda c: c.course_id, self._next_academic_period_courses)
            ),
            self.package.main_id,
        )

    def _academic_period_finished_callback(self):
        """Callback to update the state when an academic period
        has been finished.

        In this case when the academic period is finished,
        the curriculum will request for the course reports to
        the classroom.

        At first glance, this might be redundant, but this pattern
        is expected to be replicated for the multi-classroom execution mode.
        """

        report.send_message(
            report.CurriculumGeneralDebugMessage(
                "Curriculum academic period finished callback called."
            )
        )

        academic_period_records = self.classroom.get_last_courses_record()

        records_ids = set(map(lambda c: c.course_id, academic_period_records))

        courses = set(
            [
                course
                for course in self.courses
                if course.course_id in records_ids
            ]
        )

        courses_dict = {course.course_id: course for course in courses}

        for course_record in academic_period_records:

            report.send_message(
                report.CurriculumGeneralDebugMessage(
                    f"Update course state for course {course_record.course_id}"
                )
            )

            courses_dict[course_record.course_id].fetch_course_state(
                course_record
            )

        self.trigger(self.ASSIGNED_COURSES_COMPLETED_TRIGGER)

    def _curriculum_process_stop_callback(self):
        """Callback to update the state when an academic period
        has been stopped.

        It allows to change state from associated classes.

        In this case, the classroom stopped the processing, and
        the callback will restore curriculum state to READY.
        """
        self.machine.set_state(self.STATE_READY)

    def load(self):
        """Loads the Curriculum and all the Package information.

        When this method is called, the Curriculum can transition to a
        'ready' state.
        """
        super().load()

        report.send_message(
            report.CurriculumGeneralInfoMessage("Classroom initialization.")
        )

        self.classroom = SingleClassroom(
            self._mammut_session_context,
            self,
            self._academic_period_finished_callback,
            self._curriculum_process_stop_callback,
        )
        self.classroom.initialize()

        report.send_message(
            report.CurriculumGeneralInfoMessage("Curriculum load completed.")
        )

        self.trigger(self.PREPARE_TRIGGER)

    def begin(self):
        """Start the processing of the Curriculum learning.

        Use case: to start all the concurrent/parallel processing
        of lessons.

        The curriculum process will begin from the last checkpoint
        (if exists) or from the curriculum first academic period.

        Preconditions:
            - No Curriculum learning process can be started if the
                curriculum state is different than 'ready'.
        """
        if self.state == self.STATE_READY.name:
            self._started = True
            self.trigger(self.BEGIN_COURSES_TRIGGER)

        else:
            message = "Curriculum isn't in a valid state for begin(). Nothing happens."
            self.errors += message
            report.send_message(
                report.CurriculumGeneralWarningMessage(message)
            )

    def stop(self):
        """Stops the processing of the Curriculum learning.

        Use case: Stop the processing and free the computing
        resources.

        Note: Currently, calling this method does not guarantee the checkpoint
        persistence of the current lessons models being processed.
        """
        if self.state == self.STATE_ASSIGNED.name:
            self.classroom.stop()



from mammut.curriculum.lesson import lesson_base
from mammut.curriculum.models.online_models.ktt_interpretation_model import (
    KttInterpretationModel,
)
from mammut.curriculum.models.online_models.online_models import (
    OnlineModelsRegistry,
)
from mammut.curriculum.models.model_base import ModelBase
from mammut.common.corpus.corpus_map import CorpusMap
from mammut.curriculum.core.mammut_session_context import MammutSessionContext
import ray
import logging
import mammut.curriculum.report as report

logger = logging.getLogger(__name__)


class KttInterpretationLesson(lesson_base.Lesson):
    """This KttInterpretationLesson is designed to compose a course to
    compile package corpus in the Mammut Working Memory.

    Internally handles a KttInterpretationModel actor reference.

    This lesson is designed to be the first lesson of the course 0 in academic
    period 0.
    """

    RESERVED_COURSE_ID = 0
    RESERVED_LESSON_ID = 1
    RESERVED_ACADEMIC_PERIOD = 0

    def __init__(
        self,
        lesson_id: str,
        course: str,
        order: str,
        corpus_id: str,
        name: str,
        language_competence: str,
        parameters: str,
        mammut_session_context: MammutSessionContext,
    ):
        """
        Args:
            lesson_id(str): 'id' value in datasource retrieved data.
            course(str): 'course' value in datasource retrieved data.
            order(str): 'order' value in datasource retrieved data.
            corpus_id(str): 'corpus_id' value in datasource retrieved data.
            name(str): 'name' value in datasource retrieved data.
            language_competence(str): 'language_competence' value in datasource retrieved data.
            mammut_session_context: The session context from the classroom.
        """
        super(KttInterpretationLesson, self).__init__(
            lesson_id,
            course,
            order,
            corpus_id,
            None,
            name,
            language_competence,
            parameters,
            mammut_session_context,
        )

    def _get_model_instance(self) -> ModelBase:
        online_models_registry = (
            self._mammut_session_context.online_models_registry
        )
        return OnlineModelsRegistry.get_model_instance(
            KttInterpretationModel.MODEL_NAME, online_models_registry, self
        )

    def lecture(self):
        """Nothing to lecture in this lesson.
        """
        pass

    def practice(self, mammut_session_context: MammutSessionContext):
        """Practice the interpretation of the package corpus.

        This implies the interpretation of the selected corpus.

        The model takes care of creating a new java process where
        the transducer interpretation is executed.

        The model will save be saved in the OnlineModelsRegistry.
        From there, it can be used across processes to perform operations
        in the Working Memory.

        Args:
            mammut_session_context: The session context from the classroom.
        """
        self._model.train.remote(mammut_session_context)



import logging
from mammut.common.corpus.hugging_face_corpus import HuggingFaceDatasetCorpus
from typing import Union
from transformers import PreTrainedTokenizer, PreTrainedTokenizerFast
from abc import ABC, abstractmethod
from typing import Dict

log = logging.getLogger(__name__)


class HFPreprocessBase(ABC):
    """HFPreprocessBase class models a unit of preprocessing
    and formatting HuggingFace data.

    A HFPreprocessBase encapsulates:
    - A HuggingFace tokenizer.
    - Parameters for the tokenizer.
    - Preprocessed input data.

    A HFPreprocessBase should be subclassed based on the tokenizer and task to be used.
    Each HFPreprocessBase is associated with a model.

    Each task requires a HFPreprocessBase subclass
    because the data need a specific processing.
    """

    def __init__(self, tokenizer_parameters: Dict, preprocess_parameters: Dict):
        """
        Args:
            tokenizer_parameters: Dictionary with tokenizer parameters to forward to this preprocess.
            preprocess_parameters: Dictionary with preprocess parameters.
        """
        self.tokenizer_parameters: Dict = tokenizer_parameters
        self.preprocess_parameters: Dict = preprocess_parameters
        self._tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast] = self._load_tokenizer()

    @abstractmethod
    def _load_tokenizer(self):
        """Load the tokenizer for the preprocessing phase. """
        pass

    @abstractmethod
    def tokenize_corpus(self, data: HuggingFaceDatasetCorpus):
        """Tokenize loaded dataset in method get_corpus in the LessonBase class.

        Return:
            Returns encoding data required in the model train.
        """
        pass


class TokenClassificationPreprocess(HFPreprocessBase):
    def __init__(self,
                 tokenizer_parameters: Dict,
                 preprocess_parameters: Dict,
                 task: str):
        """
        Args:
            tokenizer_parameters: Dictionary with tokenizer parameters to forward to this preprocess.
            task: task specification in token classification model. i.e.: "ner", "pos" or "chunk".
        """
        self.task = task
        HFPreprocessBase.__init__(self, tokenizer_parameters, preprocess_parameters)

    def _tokenize_and_align_labels(self, examples):
        """Tokenizes the dataset. And aligns the labels with the word_ids in the tokenization results.
        In special tokens have a word id that is None. We set the label to
         -100 so they are automatically ignored in the loss function.
         We set the label for the first token of each word.
         For the other tokens in a word, we set the label to either
         the current label or -100, depending on the label_all_tokens flag.
            Args:
                examples: HuggingFace Dataset for the training.
                """
        tokenized_inputs = self._tokenizer(
            examples["tokens"], **self.preprocess_parameters
        )
        label_all_tokens = True
        labels = []
        for i, label in enumerate(examples[f"{self.task}_tags"]):
            word_ids = tokenized_inputs.word_ids(batch_index=i)
            previous_word_idx = None
            label_ids = []
            for word_idx in word_ids:
                if word_idx is None:
                    label_ids.append(-100)
                elif word_idx is not None:
                    label_ids.append(label[word_idx])
                else:
                    label_ids.append(
                        label[word_idx] if label_all_tokens else -100
                    )
                    previous_word_idx = word_idx

            labels.append(label_ids)

        tokenized_inputs["labels"] = labels
        return tokenized_inputs

    def tokenize_corpus(self, examples):
        tokenized_data = examples.map(
            self._tokenize_and_align_labels, batched=True
        )
        return tokenized_data


class HFPreprocessForQuestionAnswering(HFPreprocessBase):
    """ This subclass encapsulates the preprocessing of a question-answering task.

    """

    def __init__(
            self,
            tokenizer_parameters: Dict,
            preprocess_parameters: Dict
    ):
        HFPreprocessBase.__init__(
            self,
            tokenizer_parameters,
            preprocess_parameters
        )

    def _token_function(self, dataset):
        """
        Tokenize loaded dataset in method get_corpus
        in the LessonBase class.

        Args:
            dataset: non-preprocessed huggingface dataset.

        Return:
            Returns encoding data required in the model train.
        """

        pad_on_right = (
            self._tokenizer.padding_side
            == self.tokenizer_parameters["padding_side"]
        )

        encodings = self._tokenizer(
            dataset["question" if pad_on_right else "context"],
            dataset["context" if pad_on_right else "question"],
            truncation="only_second" if pad_on_right else "only_first",
            max_length=self.preprocess_parameters["max_length"],
            stride=self.preprocess_parameters["stride"],
            return_overflowing_tokens=True,
            return_offsets_mapping=True,
            padding="max_length",
        )
        return encodings

    def _generate_answer_end(self, answers, contexts):
        """ This method checks that the feature
        return tokenized_data answer_start is correct and
        it generates the 'answer_end' string value for the 'answers' feature.

        Args:
            answers: dataset list of answer values.
            contexts: dataset list of context values

        Return:
            Returns dataset with the new string value 'answer_end'.
        """

        for answer, context in zip(answers, contexts):
            gold_text = answer["text"][0]
            start_idx = answer["answer_start"][0]
            end_idx = start_idx + len(gold_text)

            if context[start_idx:end_idx] == gold_text:
                answer["answer_end"] = end_idx
            elif context[start_idx - 1 : end_idx - 1] == gold_text:
                answer["answer_start"] = start_idx - 1
                answer["answer_end"] = (
                    end_idx - 1
                )  # When the gold label is off by one character
            elif context[start_idx - 2 : end_idx - 2] == gold_text:
                answer["answer_start"] = start_idx - 2
                answer["answer_end"] = (
                    end_idx - 2
                )  # When the gold label is off by two characters
        return answers

    def _preprocessing(self, dataset):

        """ The method adds two values to the encoding data:
        'start_positions' and 'end_positions'.
        These values indicate the beginning and the end
        of the answer located in the value 'context'.

        Args:
            dataset: non-tokenized data.

        Return:
            Returns encoding data with the features 'start_positions'
            and 'end_positions'.
        """

        pad_on_right = (
            self._tokenizer.padding_side
            == self.tokenizer_parameters["padding_side"]
        )

        encodings = self._tokenizer(
            dataset["question" if pad_on_right else "context"],
            dataset["context" if pad_on_right else "question"],
            truncation="only_second" if pad_on_right else "only_first",
            max_length=self.preprocess_parameters["max_length"],
            stride=self.preprocess_parameters["stride"],
            return_overflowing_tokens=True,
            return_offsets_mapping=True,
            padding="max_length",
        )

        sample_mapping = encodings.pop("overflow_to_sample_mapping")
        offset_mapping = encodings.pop("offset_mapping")

        encodings["start_positions"] = []
        encodings["end_positions"] = []

        for i, offsets in enumerate(offset_mapping):
            input_ids = encodings["input_ids"][i]
            cls_index = input_ids.index(self._tokenizer.cls_token_id)
            sequence_ids = encodings.sequence_ids(i)
            sample_index = sample_mapping[i]
            answers_index = dataset["answers"][sample_index]

            if len(answers_index["answer_start"]) == 0:
                encodings["start_positions"].append(cls_index)
                encodings["end_positions"].append(cls_index)

            else:
                start_char = answers_index["answer_start"][0]
                end_char = start_char + len(answers_index["text"][0])

                token_start_index = 0
                while sequence_ids[token_start_index] != (1 if pad_on_right else 0):
                    token_start_index += 1

                token_end_index = len(input_ids) - 1
                while sequence_ids[token_end_index] != (1 if pad_on_right else 0):
                    token_end_index -= 1

                if not (offsets[token_start_index][0] <= start_char and offsets[token_end_index][1] >= end_char):
                    encodings["start_positions"].append(cls_index)
                    encodings["end_positions"].append(cls_index)
                else:
                    while token_start_index < len(offsets) and offsets[token_start_index][0] <= start_char:
                        token_start_index += 1
                    encodings["start_positions"].append(token_start_index - 1)
                    while offsets[token_end_index][1] >= end_char:
                        token_end_index -= 1
                    encodings["end_positions"].append(token_end_index + 1)

        return encodings

    def tokenize_corpus(self, dataset):
        """This method tokenizes some features of the loaded dataset in method get_corpus in the LessonBase class.

        Return:
            Returns encoding data required in the model train.
        """
        if isinstance(dataset, list):
            """This if-else block allows to preprocess a list of
            dataset objects resulting from the selection of 
            source data splits at the time of its downloading.
            """
            preprocessed_data = []
            for s in dataset:
                tokenized_data = s.map(
                    self._preprocessing,
                    batched=True,
                )
                preprocessed_data.append(tokenized_data)
            return preprocessed_data
        else:
            selected_split = self.preprocess_parameters["split"]
            preprocessed_data = dataset.map(
                self._preprocessing,
                batched=True,
                remove_columns=dataset[selected_split].column_names
            )
            return preprocessed_data


class HFPreprocessForParaphraseClassification(HFPreprocessBase):
    """ This subclass encapsulates the preprocessing of a sequence classification task.

    """

    def __init__(self, tokenizer_parameters: Dict, preprocess_parameters: Dict):
        HFPreprocessBase.__init__(
            self,
            tokenizer_parameters,
            preprocess_parameters,
        )


    @abstractmethod
    def _load_tokenizer(self):
        """Load the tokenizer for the Sequence Classification preprocessing phase. """
        pass

    def _token_function(self, dataset):
        encodings = self._tokenizer(
            dataset['sentence1'],
            dataset['sentence2'],
            dataset['label'],
            padding=True,
            truncation=self.preprocess_parameters['truncation'],
        )
        return encodings


    def tokenize_corpus(self, dataset):
        """This method implements a mapping on the input corpus.
        Tokenize loaded dataset in method get_corpus in the LessonBase class.


        Return:
            Returns encoding data required in the model train.
        """
        if isinstance(dataset, list):
            """This if-else block allows to preprocess a list of
            dataset objects resulting from the selection of 
            source data splits at the time of its downloading.
            """
            preprocessed_data = []
            for s in dataset:
                data = s.map(self._token_function, batched=True)
                tokenized_data = data.map(
                    lambda examples: {"labels": examples["label"]},
                    batched=True,
                )
                preprocessed_data.append(tokenized_data)
            return preprocessed_data
        else:
            tokenized_data = dataset.map(self._token_function, batched=True,)
            preprocessed_data = tokenized_data.map(
                lambda examples: {"labels": examples["label"]}, batched=True
            )
            return preprocessed_data



from abc import ABC, abstractmethod
from mammut.curriculum.course.course_state import CourseState
from mammut.curriculum.course.course import Course
from mammut.curriculum.core.curriculum_base import CurriculumBase
from mammut.curriculum.core.mammut_session_context import MammutSessionContext
import uuid

from typing import Dict, Set


class Classroom(ABC):
    """
    A classroom is a workspace where courses are processed.
    The classroom contains all the facilities
    required by a course to be processed.

    Contains a curriculum registry that can be used by the
    courses in the workspace.

    By having a curriculum, transitively has access to a
    Package data and a Linguistic Standard.
    """

    def __init__(self, mammut_session_context: MammutSessionContext):
        """
        Args:
            mammut_session_context: The session context for this classroom and the
                curriculums in it.
        """
        self._id = uuid.uuid4()
        self._curriculums: Dict[str, "Curriculum"] = {}
        """
        Last set of processed courses in this Classroom.
        """
        self._last_courses: Set[Course] = set()
        self._mammut_session_context = mammut_session_context

    def id(self):
        return self._id

    def add_curriculum(
        self,
        package_id: str = None,
        standard_spreadsheet_id: str = None,
        curriculum_reference: "Curriculum" = None,
        enable_ktt_interpretation: bool = False,
        package_id_is_spreadsheet: bool = False
    ):
        """
        Adds a new curriculum facility to this classroom.

        If no curriculum_reference is given, the curriculum is instantiated.
        """
        if curriculum_reference is not None:
            self._curriculums[
                curriculum_reference.package.main_id
            ] = curriculum_reference
        else:
            curriculum = CurriculumBase(
                package_id,
                standard_spreadsheet_id,
                package_id_is_spreadsheet=package_id_is_spreadsheet,
                show_progress=False,
                create_progress_sheet=False,
                mammut_session_context=self._mammut_session_context,
                enable_ktt_interpretation=enable_ktt_interpretation
            )
            curriculum.load()
            self._curriculums[package_id] = curriculum

    def get_last_courses_record(self) -> Set[CourseState]:
        """Return the last set of processed courses.

        By requesting last processed courses record, client
        control classes can fetch the latest data state of
        courses processed in this classroom.

        Returns:
            Set of course Records.
        """
        records = set(
            map(lambda course: course.get_course_record(), self._last_courses)
        )
        return records

    @abstractmethod
    def initialize(self):
        """
        initialize resources for a particular classroom type.
        """
        pass

    @abstractmethod
    def take_courses(
        self, courses_ids: Set[int], curriculum_spreadsheet_id: str
    ) -> None:
        """
        Process the desired courses in this classroom.

        This method does not return anything to client control classes,
        is intended to side effect the processing of courses.

        Args:
            courses_ids(Set[int]): list of courses ids to process.
            curriculum_spreadsheet_id: spreadsheet id of the curriculum
                where courses belong.

        Returns:
            A set of course records encapsulating the state of
            processed courses.
        """
        pass

    @abstractmethod
    def stop(self):
        """
        Request gracefully stopping of this classroom execution process.
        """
        pass



# coding=utf-8
from mammut.visualization.simulator_widget import *


class Simulator:
    def __init__(
        self,
        users={},
        current_user=None,
        mammut_api=None,
        package=None,
        redis_memory=None,
        rooms={},
        chat_room=None,
        current_person_user=None,
        room_thoughts=None,
        supervisor=None,
    ):
        self.users = users
        self.current_user = current_user
        self.current_person_user = current_person_user
        self.mammut_api = mammut_api
        self.package = package
        self.redis_memory = redis_memory
        self.rooms = rooms
        self.chat_room = chat_room
        self.room_thoughts = room_thoughts
        self.supervisor = supervisor
        execution_mode = os.getenv("EXECUTION-MODE", "distributed")
        if execution_mode == "distributed":
            self.embedded = False
        else:
            self.embedded = True

    def get_simulator_widget(
        self,
        mammut_api_url,
        redis_address,
        redis_port,
        package_file,
        standard_spreadsheet_id,
        users_filename,
        users_rooms_spreadsheet_id,
        users_sheet_name,
        rooms_sheet_name,
        rooms_filename,
    ):
        return get_simulator_widget(
            self,
            mammut_api_url,
            redis_address,
            redis_port,
            package_file,
            standard_spreadsheet_id,
            users_filename,
            users_rooms_spreadsheet_id,
            users_sheet_name,
            rooms_sheet_name,
            rooms_filename,
        )



# coding=utf-8

from typing import Optional, List

import mammut.assessment.util
from mammut.common.storage.storage_manager import StorageManager
import pandas as pd
import logging
from mammut.assessment.assessment import Assessment, Problem
from mammut.assessment.problem_type_enum import ProblemType
import mammut.assessment.util as assessment_utils
from typing import Dict

log = logging.getLogger(__name__)


class Assessor:
    """Class to handle assessments and hide complexity
    from clients.

    It contains the in-memory dataframe of assessments.
    And provide a public API for lazy concrete instantiation to
    client classes.
    """

    _storage_manager = StorageManager()

    def __init__(self, package_spreadsheet_id: str, spreadsheet_name):
        self.errors: List[str] = []
        self.valid: bool = True
        self.spreadsheet_id = package_spreadsheet_id
        self.spreadsheet_name = spreadsheet_name
        self._dataframe: pd.DataFrame = Optional[None]
        """
            Initializes the Problems registry for concrete instantiation of
            problems at runtime.
        """
        self._problems_registry = {}
        self._load_problem_registry()

    def load(self):
        base_dataframe = self._storage_manager.get_spreadsheet_as_dataframe(
            self.spreadsheet_id,
            self.spreadsheet_name,
            mammut.assessment.util.DEFAULT_START_COLUMN,
            mammut.assessment.util.DEFAULT_START_ROW,
            mammut.assessment.util.DEFAULT_END_CELL,
        )
        self._dataframe = base_dataframe

        """
           Validates the problems types values for the specified assessments
        """
        if not all(
            base_dataframe[mammut.assessment.util.PROBLEM_TYPE_COLUMN].isin(
                ProblemType.names()
            )
        ):
            self.valid = False
            self.errors += (
                "Invalid Problem(s) type specified in assessments sheet."
            )

    def get_course_max_score(self, course_id) -> float:
        """Returns the sum of all the problems assigned to the Course.
        """
        partial_df: pd.DataFrame = self._dataframe.loc[
            self._dataframe[mammut.assessment.util.COURSE_ID_COLUMN]
            == str(course_id)
        ]
        total_score = sum(
            map(
                lambda w: float(w),
                partial_df[mammut.assessment.util.WEIGHT_COLUMN].array,
            )
        )
        return total_score

    def get_scheduled_assessment(
        self, course_id, schedule
    ) -> Optional[Assessment]:
        """Returns an assessments for a schedule in a course.

        Args:
            course_id(int): course identifier.
            schedule(int): assessment schedule, indicates after which lesson
                should the assessment be evaluated.
            model: Models enum value required for concrete instantiation.

        Returns:
            The Assessment for the given schedule.
        """

        """
            Get only the partial dataframe for the course and lesson schedule.
        """
        partial_df: pd.DataFrame = self._dataframe.loc[
            (
                self._dataframe[mammut.assessment.util.COURSE_ID_COLUMN]
                == str(course_id)
            )
            & (
                self._dataframe[mammut.assessment.util.SCHEDULE_COLUMN]
                == str(schedule)
            )
        ]

        """
            Scheduled assessments are optionals.
        """
        if partial_df.empty:
            return None

        """
            Get the ID of the current assessment.
        """
        assessment_id = partial_df[
            mammut.assessment.util.ASSESSMENT_ID_COLUMN
        ].iloc[0]
        """
            For the assessment get the problem rows, and pass those problems
            to Assessment class constructor. 
        """
        problems = []
        for index, problem_row in partial_df.iterrows():
            problem_type_value = problem_row[
                mammut.assessment.util.PROBLEM_TYPE_COLUMN
            ]
            """
                Get the custom class from the internal problems dictionary, and 
                instantiate an object of that class.
            """
            problem: Problem = self._problems_registry[problem_type_value](
                problem_row[mammut.assessment.util.COURSE_ID_COLUMN],
                problem_row[mammut.assessment.util.ASSESSMENT_ID_COLUMN],
                problem_row[mammut.assessment.util.SCHEDULE_COLUMN],
                problem_row[mammut.assessment.util.PROBLEM_ID_COLUMN],
                problem_row[mammut.assessment.util.PROBLEM_TYPE_COLUMN],
                problem_row[mammut.assessment.util.PARAMETERS_COLUMN],
                problem_row[mammut.assessment.util.LANGUAGE_COMPETENCE_COLUMN],
                problem_row[mammut.assessment.util.WEIGHT_COLUMN],
                problem_row[mammut.assessment.util.RETRY_COLUMN],
                problem_row[mammut.assessment.util.OBSERVATIONS_COLUMN],
            )
            problems.append(problem)

        assessment: Assessment = Assessment(assessment_id, problems, course_id)
        return assessment

    def _load_problem_registry(self):
        """Loads the problems type internal dictionary used in
        concrete instantiation operations.

        The dictionary keys are strings for the problem type, and the
        values are problems custom classes.
        """
        for problem_type in ProblemType.list():
            self._problems_registry[problem_type[0]] = problem_type[1]



# coding=utf-8
from __future__ import ( division, absolute_import, print_function, unicode_literals )

import sys, os

if sys.version_info >= (3,):
    import urllib.request as urlrequest
    import urllib.parse as urlparse
    import bz2
    import gzip
else:
    import urllib2 as urlrequest
    import urlparse
    import bz2
    import gzip


def file(url, directory, file=None):
    u = urlrequest.urlopen(url)

    scheme, netloc, path, query, fragment = urlparse.urlsplit(url)

    if not file:
        file = os.path.basename(path)

    filename = os.path.join(directory, file)
    if os.path.isfile(filename):
        return filename

    with open(filename, 'wb') as f:
        meta = u.info()
        meta_func = meta.getheaders if hasattr(meta, 'getheaders') else meta.get_all
        meta_length = meta_func("Content-Length")
        file_size = None
        if meta_length:
            file_size = int(meta_length[0])
        print("Downloading: {0} Bytes: {1}".format(url, file_size))

        file_size_dl = 0
        block_sz = 8192
        while file_size_dl < file_size:
            reminder = file_size - file_size_dl
            if reminder < block_sz:
                block_sz = reminder
            buffer = u.read(block_sz)

            file_size_dl += len(buffer)
            f.write(buffer)

            status = "{0:16}".format(file_size_dl)
            if file_size:
                status += "   [{0:6.2f}%]".format(file_size_dl * 100 / file_size)
            status += chr(13)
            print(status, end="")
        print()

    return filename

def decompress_bz2(compressed_file, decompressed_file=None):
    if not decompressed_file:
        decompressed_file = os.path.splitext(compressed_file)[0]
    if os.path.isfile(decompressed_file):
        return decompressed_file
    with open(decompressed_file, 'wb') as new_file, bz2.BZ2File(compressed_file, 'rb') as file:
        for data in iter(lambda: file.read(100 * 1024), b''):
            new_file.write(data)

    return decompressed_file

def decompress_gz(compressed_file, decompressed_file=None):
    if not decompressed_file:
        decompressed_file = os.path.splitext(compressed_file)[0]
    if os.path.isfile(decompressed_file):
        return decompressed_file
    with open(decompressed_file, 'wb') as new_file, gzip.open(compressed_file, 'rb') as file:
        for data in iter(lambda: file.read(100 * 1024), b''):
            new_file.write(data)

    return decompressed_file


from typing import List

from mammut.common.synthetic.synthetic_corpus_generator.morphemes_generator import (
    MorphemesGenerator,
)
import random
import string


def generate_n_random_free(n: int, ext: int) -> List[str]:
    """Generates n unique random strings.

    TODO: Formalize the upper limit of n.

    Args:
        n(int): The number of strings to generate.
        ext(int): The number of chars in the random
            strings.

    Returns:
        The list of generated random strings
    """

    def get_random_char():
        return random.choice(string.ascii_letters)

    def get_random_string():
        s = list(get_random_char())
        for j in range(0, ext - 1):
            t_s = s + list(get_random_char())
            c = 0
            while "".join(t_s) in generated_count:
                t_s = s + list(get_random_char())
                c += 1
                if c == 100:
                    break

            s = t_s
            generated_count.add("".join(s))
        return "".join(s)

    generated_count = set()
    ret = []
    for i in range(0, n):
        ret.append(get_random_string())
    return ret


class SyntheticCorpusGenerator:
    """Synthetic corpus generation class
    Encapsulates the algorithm and considerations
    for the retrieval of a synthetic corpus:
    - Substitution rules for free morphemes character
    - Accountability of frequencies
    - Handling of distributions for each type of
        synthetic corpus
    """

    def __init__(self, morpheme_generator: MorphemesGenerator):
        self.morphemes_generator = morpheme_generator

    def generate_synthetic_corpus(
        self,
        nouns_words: int,
        adjectives_words: int,
        verbs_words: int,
        nouns_bound: int,
        adjectives_bound: int,
        verbs_bound: int,
        free_lemma_words: int,
        random_ext: int = 3
    ) -> List[str]:

        if random_ext < 3:
            raise Exception("Random string length too short. Introduce a value >= 3")

        free_lemma_words_arr = []
        nouns_words_arr = []
        adjectives_words_arr = []
        verbs_words_arr = []
        nouns_bound_arr = []
        adjectives_bound_arr = []
        verbs_bound_arr = []

        if free_lemma_words >= 1:
            free_lemma_words_arr = [m.free for m in self.morphemes_generator.get_free_lemma_from_dictionary() if not m.mixed]
            free_lemma_words_arr = list(set(map(
                lambda morpheme_paradigm: morpheme_paradigm, free_lemma_words_arr
            )))
            for i in range(0, free_lemma_words):
                free_lemma_words_arr = free_lemma_words_arr + free_lemma_words_arr

        if nouns_words >= 1:
            nouns_words_arr = [
                m
                for m in self.morphemes_generator.get_tokens("NOUN", "es")
                if not m.mixed
            ]
            for i in range(0, nouns_words):
                nouns_words_arr = nouns_words_arr + [
                    m
                    for m in self.morphemes_generator.get_tokens("NOUN", "es")
                    if not m.mixed
                ]

        if adjectives_words >= 1:
            adjectives_words_arr = [
                m
                for m in self.morphemes_generator.get_tokens("ADJ", "es")
                if not m.mixed
            ]
            for i in range(0, adjectives_words):
                adjectives_words_arr = adjectives_words_arr + [
                    m
                    for m in self.morphemes_generator.get_tokens("ADJ", "es")
                    if not m.mixed
                ]

        if verbs_words >= 1:
            verbs_words_arr = [
                m
                for m in self.morphemes_generator.get_tokens("VERB", "es")
                if not m.mixed
            ]
            for i in range(0, verbs_words):
                verbs_words_arr = verbs_words_arr + [
                    m
                    for m in self.morphemes_generator.get_tokens("VERB", "es")
                    if not m.mixed
                ]

        if nouns_bound >= 1:
            nouns_bound_arr = list(
                set(
                    [
                        m.bound
                        for m in self.morphemes_generator.get_tokens(
                            "NOUN", "es"
                        )
                        if not m.mixed
                    ]
                )
            )
            for i in range(0, nouns_bound):
                nouns_bound_arr = nouns_bound_arr + list(
                    set(
                        [
                            m.bound
                            for m in self.morphemes_generator.get_tokens(
                                "NOUN", "es"
                            )
                            if not m.mixed
                        ]
                    )
                )

        if adjectives_bound >= 1:
            adjectives_bound_arr = list(
                set(
                    [
                        m.bound
                        for m in self.morphemes_generator.get_tokens(
                            "ADJ", "es"
                        )
                        if not m.mixed
                    ]
                )
            )
            for i in range(0, adjectives_bound):
                adjectives_bound_arr = adjectives_bound_arr + list(
                    set(
                        [
                            m.bound
                            for m in self.morphemes_generator.get_tokens(
                                "ADJ", "es"
                            )
                            if not m.mixed
                        ]
                    )
                )

        if verbs_bound >= 1:
            verbs_bound_arr = list(
                set(
                    [
                        m.bound
                        for m in self.morphemes_generator.get_tokens(
                            "VERB", "es"
                        )
                        if not m.mixed
                    ]
                )
            )
            for i in range(0, verbs_bound):
                verbs_bound_arr = verbs_bound_arr + list(
                    set(
                        [
                            m.bound
                            for m in self.morphemes_generator.get_tokens(
                                "VERB", "es"
                            )
                            if not m.mixed
                        ]
                    )
                )

        return self._generate_tokens(
            nouns_words_arr,
            adjectives_words_arr,
            verbs_words_arr,
            nouns_bound_arr,
            adjectives_bound_arr,
            verbs_bound_arr,
            free_lemma_words_arr,
            random_ext
        )

    def _generate_tokens(
        self,
        nouns_words,
        adjectives_words,
        verbs_words,
        nouns_bound: List[str],
        adjectives_bound: List[str],
        verbs_bound: List[str],
        free_lemmas_words: List[str],
        random_ext: int
    ) -> List[str]:
        ret_list = nouns_words + adjectives_words + verbs_words
        bound_morphemes_set = set([m.bound for m in nouns_words + adjectives_words + verbs_words])
        random_frees = generate_n_random_free(len(ret_list), random_ext)
        free_dict = dict()
        bound_dict = dict()
        ret_list_free_altered = []

        for morpheme in ret_list:
            free_morpheme = random_frees.pop()
            if free_morpheme in free_dict:
                free_dict[free_morpheme] += 1
            else:
                free_dict[free_morpheme] = 1

            if morpheme.bound in bound_dict:
                bound_dict[morpheme.bound] += 1
            else:
                bound_dict[morpheme.bound] = 1

            ret_list_free_altered.append(free_morpheme + morpheme.bound)

        free_lemmas_words = list(free_lemmas_words)
        random_bound = generate_n_random_free(len(free_lemmas_words), random_ext)
        free_lemmas_words_random_suffix = list(
            map(lambda f_w: f_w + random_bound.pop(), free_lemmas_words)
        )

        unsorted_tokens = (
            ret_list_free_altered
            + nouns_bound
            + adjectives_bound
            + verbs_bound
            + free_lemmas_words_random_suffix
        )

        for i in nouns_bound:
            bound_morphemes_set.add(i)
        for i in adjectives_bound:
            bound_morphemes_set.add(i)
        for i in verbs_bound:
            bound_morphemes_set.add(i)

        random.shuffle(unsorted_tokens)
        return list(free_lemmas_words) + list(bound_morphemes_set) + list(unsorted_tokens)



# coding=utf-8
from kafka import KafkaConsumer, KafkaProducer
import logging
from mammut.apps import InterpreterMain_pb2

from mammut.apps import (
    KAFKA_BROKER_STR,
    COMPILER_SERVICE_INPUT_TOPIC,
    COMPILER_SERVICE_OUTPUT_TOPIC,
    COMPILER_OUTPUT_RETRIES,
    COMPILER_OUTPUT_TIMEOUT,
    COMPILER_OUTPUT_MAX_RECORDS,
)
from mammut.common.util import StringEnum

log = logging.getLogger(__name__)

FINAL_OUTPUT_MESSAGE_PREFIX = "Finished compilation for Compilation Task: {}"


class CorpusCompilerServiceCallResult(StringEnum):
    NoneCall = "NoneCall"
    Unknown = "Unknown"
    FailedCall = "FailedCall"
    SuccessCall = "SuccessCall"


def is_error_message(imo: InterpreterMain_pb2.InterpreterMainOutput):

    imo_type = imo.type
    switcher = {
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "SuccessOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "FailureOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "UnsupportedFeatureOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "PropertyValueNotFoundOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "UnknownPropertyOrEdgeOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "OntologyTypeNotFoundOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "VertexUndefinedOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "VariableUndefinedOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "FieldUndefinedOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "EdgeValueNotFoundOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "ReservedKeywordFoundOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "PropertyDescriptorNotFoundOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "GettingCorpusMapOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "GettingOntologyOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "ExpandingOntologyOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "ComputingEntryPointOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "GettingOntologyInstancesOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "LoadingInstancesToWorkingMemoryOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "GeneratingOntologyClassesOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "GettingVariablesOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "GettingSceneriesOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "GettingExtensionsOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "CompilingVariablesOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "CompilingSceneriesOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "CompilingExtensionsOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "GroupingCompiledExtensionsOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "GroupingCompiledSceneriesOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "CompilingInformationTuplesOutput"
        ].number: True,
        InterpreterMain_pb2.InterpreterMainOutput.Types.DESCRIPTOR.values_by_name[
            "UnexpectedCompilerError"
        ].number: True,
    }
    if switcher.__contains__(imo_type):
        return switcher[imo_type]
    else:
        return False


class CorpusCompilerService:
    def __init__(self):
        self.producer = KafkaProducer(bootstrap_servers=[KAFKA_BROKER_STR], retries=5)
        self.consumer = KafkaConsumer(
            COMPILER_SERVICE_OUTPUT_TOPIC,
            enable_auto_commit=True,
            group_id="framework",
            bootstrap_servers=[KAFKA_BROKER_STR],
            api_version=(0, 10),
        )

    def compile(
        self,
        mammut_id: int,
        slide_id: str,
        execute_sampling_corpus: bool,
        regional_settings: str,
        metadata_token_separator: str,
        corpus_ids: [],
        compilation_message_callback,
    ):
        raw_messages = self.consumer.poll(
            timeout_ms=COMPILER_OUTPUT_TIMEOUT, max_records=COMPILER_OUTPUT_MAX_RECORDS
        )
        for topic_partition, messages in raw_messages.items():
            for message in messages:
                output = InterpreterMain_pb2.InterpreterMainOutput()
                output.ParseFromString(message.value)
                mess = "Previous unread message found: {}".format(output.message)
                log.info(mess)
                if compilation_message_callback:
                    compilation_message_callback(mess, None)

        def on_send_success(record_metadata):
            mess = "Message delivered to {} [{} - {}]".format(
                record_metadata.topic, record_metadata.partition, record_metadata.offset
            )
            log.info(mess)
            if compilation_message_callback:
                compilation_message_callback(mess, None)

        def on_send_error(excp):
            mess = "Compilation Error: Message delivery failed: {}".format(excp)
            if compilation_message_callback:
                compilation_message_callback(mess, None)
            log.error(mess, exc_info=excp)

        input = InterpreterMain_pb2.InterpreterMainInput()
        input.regionalSettings = regional_settings
        input.mammutId = mammut_id
        input.slideId = slide_id
        input.executeSamplingCorpus = execute_sampling_corpus
        input.corpusIds.extend(corpus_ids)
        input.metadataTokenSeparator = metadata_token_separator
        self.producer.send(
            COMPILER_SERVICE_INPUT_TOPIC,
            key=str(mammut_id).encode(),
            value=input.SerializeToString(),
        ).add_callback(on_send_success).add_errback(on_send_error)
        # block until all async messages are sent
        self.producer.flush()
        outputs = []
        retries = COMPILER_OUTPUT_RETRIES
        while retries > 0:
            retries -= 1
            raw_messages = self.consumer.poll(
                timeout_ms=COMPILER_OUTPUT_TIMEOUT,
                max_records=COMPILER_OUTPUT_MAX_RECORDS,
            )
            for topic_partition, messages in raw_messages.items():
                for message in messages:
                    output = InterpreterMain_pb2.InterpreterMainOutput()
                    output.ParseFromString(message.value)
                    outputStr = "Additional info: Phase->" + str(output.phase)
                    if output.mammutId == mammut_id:
                        retries = COMPILER_OUTPUT_RETRIES
                        outputs.append(output)
                        if compilation_message_callback and is_error_message(output):
                            compilation_message_callback(output.message, outputStr)

                        log.info(output.message)
                        if output.message.startswith(
                            FINAL_OUTPUT_MESSAGE_PREFIX.format(mammut_id)
                        ):
                            retries = 0
                    else:
                        if compilation_message_callback and is_error_message(output):
                            compilation_message_callback(output.message, outputStr)

                        log.info(output.message)
        return outputs



import click
import os


def get_version_number():
    path = os.path.dirname(__file__)
    file_name = path + "/../../../pyproject.toml"
    setup_file = open(file_name, 'r')
    lines = setup_file.read().split('\n')
    version_number_range=[]
    for index, line in enumerate(lines):
        if 'version' in line:
            for letter_index, letter in enumerate(line):
                if letter == '"':
                    version_number_range.append(letter_index)
            return line[version_number_range[0]+1:version_number_range[1]]

@click.command()
@click.version_option(get_version_number())
def main():
    """mammutctl script."""
    click.echo('mammutctl under construction!')



from mammut.curriculum.classroom.classroom import Classroom
import threading
from typing import Callable, Set, Optional
from mammut.curriculum.core.mammut_session_context import MammutSessionContext


class SingleClassroom(Classroom):
    """
        Single classroom designed for academic period courses
        execution on a single machine.
    """

    def __init__(
        self,
        mammut_session_context: MammutSessionContext,
        curriculum: Optional["Curriculum"] = None,
        academic_period_finished_callback: Optional[Callable] = None,
        curriculum_process_stopped_callback: Optional[Callable] = None,
    ):
        """
        Arguments are optional for subclassing convenience.

        Args:
            mammut_session_context: The session context for this classroom and the
                curriculums in it.
            curriculum: reference for single machine shared curriculum.
            academic_period_finished_callback: callback to notify academic period
                finished.
            curriculum_process_stopped_callback: callback to notify processing stop.
        """
        super().__init__(mammut_session_context)
        if curriculum is not None:
            self.add_curriculum(curriculum_reference=curriculum)

        self._academic_period_finished_callback = (
            academic_period_finished_callback
        )
        self._process_stopped_callback = curriculum_process_stopped_callback

        """
        Classroom processing is performed sequentially in a single thread.
        """
        self._main_thread: threading.Thread = None
        self._stop_main_thread_event: threading.Event = None

    def initialize(self):
        self._stop_main_thread_event = threading.Event()

    def _process_courses_loop(
        self, courses_ids: Set[int], curriculum_spreadsheet_id: str
    ):
        """
        Process courses found in courses_ids. This method communicates with
        Curriculum instance for state changing using provided callbacks.
        """
        if self._stop_main_thread_event.is_set():
            """
            The next academic period wouldn't be processed,
            since Curriculum requested stop. 
            """
            self._process_stopped_callback()
            self._stop_main_thread_event.clear()
            return

        curriculum = self._curriculums[curriculum_spreadsheet_id]
        self._last_courses = set(
            [
                course
                for course in curriculum.courses
                if course.course_id in courses_ids
            ]
        )

        for c in self._last_courses:
            if self._stop_main_thread_event.is_set():
                """
                Curriculum requested stop in the middle of an
                academic period. 
                """
                self._process_stopped_callback()
                self._stop_main_thread_event.clear()
                return
            c.start(curriculum.progress, self._mammut_session_context)

        """
        The academic period finished callback will be invoked 
        in new thread. Doing so, the current thread 
        in charge of current academic period processing will reach 
        it's end.

        The new thread will invoke the callback, it will cause 
        the curriculum to compute next semester, and finally will 
        call the take_courses method again thanks to state transitions.
        """
        transition_thread = threading.Thread(
            target=self._academic_period_finished_callback,
            name="academic_period_finished_transition",
        )
        transition_thread.start()

    def take_courses(
        self, courses_ids: Set[int], curriculum_spreadsheet_id: str
    ):
        self._main_thread = threading.Thread(
            target=self._process_courses_loop,
            args=(courses_ids, curriculum_spreadsheet_id),
            name="academic_period_processing",
        )
        self._main_thread.start()

    def stop(self):
        self._stop_main_thread_event.set()



# coding=utf-8

from mammut.curriculum.lesson import lesson_base
from mammut.common.corpus.corpus_map import CorpusMap
from mammut.curriculum.models import mock
from mammut.curriculum.core.mammut_session_context import MammutSessionContext


class MockLesson(lesson_base.Lesson):
    """Implements a Lesson mock for unit testing of Curriculum.
    """

    def __init__(
        self,
        lesson_id: str,
        course: str,
        order: str,
        corpus_id: str,
        corpus_map: CorpusMap,
        name: str,
        language_competence: str,
        parameters: str,
        mammut_session_context: MammutSessionContext,
    ):
        lesson_base.Lesson.__init__(
            self,
            lesson_id,
            course,
            order,
            corpus_id,
            corpus_map,
            name,
            language_competence,
            parameters,
            mammut_session_context,
        )
        self._model = self._get_model_instance()

    def lecture(self):
        pass

    def practice(self, mammut_session_context: MammutSessionContext):
        """
        Args:
            save_path: file system folder path to save the model after trained.
        """
        self._model.train(mammut_session_context)
        self._model.save(mammut_session_context)

    def _get_model_instance(self):
        return mock.TestMockModel(self.course, self.lesson_id)



# coding=utf-8
from ipywidgets import *
from django.template.loader import get_template
import pandas as pd
import plotly
import plotly.graph_objs as go
from collections import namedtuple

rae_verb_conjugation_models_template = get_template("rae_verb_conjugation_models.html")
pending_words_template = get_template("pending_words.html")
paradigm_application_template = get_template("paradigm_application.html")
ParadigmAppWordDesc = namedtuple("ParadigmAppWordDesc", ["morpheme", "features"])
PendingWordDesc = namedtuple("PendingWordDesc", ["text", "strike"])


def create_conjugator_section(verb: str, language_tool: str = "es-wr"):
    res = None
    if "en" in language_tool:
        verbix_style = ""  #'border: 0px none; margin-left: -132px; height: 650px; margin-top: -230px; width: 900px; margin-bottom: -170px;'
        verbix_url_pattern = "//www.verbix.com/webverbix/English/{0}.html"
        verbix_url = verbix_url_pattern.format(verb.lower().strip())
        if verbix_style:
            verbix_html_wrapper = """<div style="border: 2px solid #D5CC5A; overflow: hidden; margin: 15px auto; max-width: 950px;">
        <iframe src="{0}" style="{1}">
        </iframe>
        </div>"""
            verbix_html = verbix_html_wrapper.format(verbix_url, verbix_style)
        else:
            verbix_html_wrapper = '<iframe src="{0}" width=750 height=350></iframe>'
            verbix_html = verbix_html_wrapper.format(verbix_url)
        res = widgets.HTML(value=verbix_html)
    else:
        word_reference_style = "border: 0px none; margin-left: -132px; height: 650px; margin-top: -230px; width: 900px; margin-bottom: -170px;"
        word_reference_url_pattern = "//www.wordreference.com/conj/EsVerbs.aspx?v={0}"
        word_reference_url = word_reference_url_pattern.format(verb)
        if word_reference_style:
            word_reference_html_wrapper = """<div style="border: 2px solid #D5CC5A; overflow: hidden; margin: 15px auto; max-width: 950px;">
    <iframe src="{0}" style="{1}">
    </iframe>
    </div>"""
            word_reference_html = word_reference_html_wrapper.format(
                word_reference_url, word_reference_style
            )
        else:
            word_reference_html_wrapper = (
                '<iframe src="{0}" width=750 height=350></iframe>'
            )
            word_reference_html = word_reference_html_wrapper.format(word_reference_url)
        res = widgets.HTML(value=word_reference_html)
    return res


def create_rae_conjugation_section():
    # rae_url = 'http://www.rae.es/diccionario-panhispanico-de-dudas/apendices/modelos-de-conjugacion-verbal'
    rae_style = "border: 0px none; margin-left: -132px; height: 650px; margin-top: -230px; width: 900px; margin-bottom: -170px;"
    if rae_style:
        rae_html_wrapper = """<div style="border: 2px solid #D5CC5A; overflow: hidden; margin: 15px auto; max-width: 950px;">
    <iframe style="{1}">{0}</iframe></div>"""
        rae_html = rae_html_wrapper.format(
            rae_verb_conjugation_models_template.render(), rae_style
        )
    else:
        rae_html_wrapper = "<iframe width=750 height=350>{0}</iframe>"
        rae_html = rae_html_wrapper.format(
            rae_verb_conjugation_models_template.render()
        )
    rae = widgets.HTML(value=rae_verb_conjugation_models_template.render())
    box = widgets.Accordion([rae])
    box.set_title(0, "Modelos de conjugacion de RAE")
    return box


def create_defined_words_section(words, as_data_frame=True):
    buckets = {}
    max_bucket_size = 0
    for word in words:
        bucket = []
        if word[0] not in buckets:
            buckets[word[0]] = bucket
        else:
            bucket = buckets[word[0]]
        bucket.append(word)
        if len(bucket) > max_bucket_size:
            max_bucket_size = len(bucket)
    for bid in buckets:
        buckets[bid] = sorted(buckets[bid])
    bids = sorted(buckets.keys())
    word_groups_rows = []
    for i in range(0, max_bucket_size):
        row = []
        word_groups_rows.append(row)
        for bkt_id in bids:
            bucket = buckets[bkt_id]
            if len(bucket) > i:
                if as_data_frame:
                    row.append(bucket[i])
                else:
                    row.append(PendingWordDesc(bucket[i], False))
            else:
                if as_data_frame:
                    row.append("")
                else:
                    row.append(PendingWordDesc("-", False))
    if as_data_frame:
        return pd.DataFrame(data=word_groups_rows, columns=bids)
    else:
        context_dict = {"word_groups": bids, "word_groups_rows": word_groups_rows}
        pending_words_html = pending_words_template.render(context_dict)
        return widgets.HTML(value=pending_words_html)


def create_pending_words_section(
    dictionary, regional_settings: str, layer=0, as_data_frame=True
):
    return create_words_section(
        dictionary, regional_settings, layer, as_data_frame, True
    )


def create_ready_words_section(
    dictionary, regional_settings: str, layer=0, as_data_frame=True
):
    return create_words_section(
        dictionary, regional_settings, layer, as_data_frame, False
    )


def create_words_section(
    dictionary, regional_settings: str, layer, as_data_frame, undefined
):
    buckets = {}
    max_bucket_size = 0
    word_list = (
        dictionary.get_undefined_morphemes_generator(regional_settings, layer, True)
        if undefined
        else dictionary.get_defined_morphemes_generator(regional_settings, layer, True)
    )
    for word, index in word_list:
        bucket = []
        if word[0] not in buckets:
            buckets[word[0]] = bucket
        else:
            bucket = buckets[word[0]]
        bucket.append("{0} | {1:.2f}".format(word, index))
        if len(bucket) > max_bucket_size:
            max_bucket_size = len(bucket)
    for bid in buckets:
        buckets[bid] = sorted(buckets[bid])
    bids = sorted(buckets.keys())
    word_groups_rows = []
    for i in range(0, max_bucket_size):
        row = []
        word_groups_rows.append(row)
        for bkt_id in bids:
            bucket = buckets[bkt_id]
            if len(bucket) > i:
                if as_data_frame:
                    row.append(bucket[i])
                else:
                    row.append(PendingWordDesc(bucket[i], False))
            else:
                if as_data_frame:
                    row.append("")
                else:
                    row.append(PendingWordDesc("-", False))
    if as_data_frame:
        return pd.DataFrame(data=word_groups_rows, columns=bids)
    else:
        context_dict = {"word_groups": bids, "word_groups_rows": word_groups_rows}
        pending_words_html = pending_words_template.render(context_dict)
        return widgets.HTML(value=pending_words_html)


def create_undefined_functionals_section(
    standard, regional_settings: str, as_data_frame=True
):
    buckets = {}
    max_bucket_size = 0
    for func in standard.get_undefined_functionals(regional_settings):
        word = func.lemma + " (" + func.pos_paradigm + ")"
        bucket = []
        if word[0] not in buckets:
            buckets[word[0]] = bucket
        else:
            bucket = buckets[word[0]]
        bucket.append(word)
        if len(bucket) > max_bucket_size:
            max_bucket_size = len(bucket)
    for bid in buckets:
        buckets[bid] = sorted(buckets[bid])
    bids = sorted(buckets.keys())
    word_groups_rows = []
    for i in range(0, max_bucket_size):
        row = []
        word_groups_rows.append(row)
        for bkt_id in bids:
            bucket = buckets[bkt_id]
            if len(bucket) > i:
                if as_data_frame:
                    row.append(bucket[i])
                else:
                    row.append(PendingWordDesc(bucket[i], False))
            else:
                if as_data_frame:
                    row.append("")
                else:
                    row.append(PendingWordDesc("-", False))
    if as_data_frame:
        return pd.DataFrame(data=word_groups_rows, columns=bids)
    else:
        context_dict = {"word_groups": bids, "word_groups_rows": word_groups_rows}
        pending_words_html = pending_words_template.render(context_dict)
        return widgets.HTML(value=pending_words_html)


def create_words_in_definitions_out_of_vocabulary_section(
    dictionary, regional_settings: str, as_data_frame=True
):
    buckets = {}
    max_bucket_size = 0
    for word in dictionary.get_words_in_definitions_out_of_vocabulary(
        regional_settings
    ):
        bucket = []
        if word[0] not in buckets:
            buckets[word[0]] = bucket
        else:
            bucket = buckets[word[0]]
        bucket.append(word)
        if len(bucket) > max_bucket_size:
            max_bucket_size = len(bucket)
    for bid in buckets:
        buckets[bid] = sorted(buckets[bid])
    bids = sorted(buckets.keys())
    word_groups_rows = []
    for i in range(0, max_bucket_size):
        row = []
        word_groups_rows.append(row)
        for bkt_id in bids:
            bucket = buckets[bkt_id]
            if len(bucket) > i:
                if as_data_frame:
                    row.append(bucket[i])
                else:
                    row.append(PendingWordDesc(bucket[i], False))
            else:
                if as_data_frame:
                    row.append("")
                else:
                    row.append(PendingWordDesc("-", False))
    if as_data_frame:
        return pd.DataFrame(data=word_groups_rows, columns=bids)
    else:
        context_dict = {"word_groups": bids, "word_groups_rows": word_groups_rows}
        pending_words_html = pending_words_template.render(context_dict)
        return widgets.HTML(value=pending_words_html)


def show_lexicalization_evolution(tokenizer, regional_settings: str):
    (
        plot_data_words_count,
        plot_data_words_count_colors,
        plot_data_undefined_morphemes_count,
        plot_data_undefined_morphemes_count_colors,
    ) = tokenizer.stats[regional_settings].get_plot_data()
    trace_words_count = go.Scatter(
        y=plot_data_words_count,
        mode="markers+lines",
        name="Cantidad de Palabras",
        marker=dict(color=plot_data_words_count_colors),
    )
    trace_plot_data_undefined_morphemes_count = go.Scatter(
        y=plot_data_undefined_morphemes_count,
        mode="markers+lines",
        name="Palabras sin definicion",
        marker=dict(color=plot_data_undefined_morphemes_count_colors),
    )
    data = [trace_words_count, trace_plot_data_undefined_morphemes_count]

    plotly.offline.iplot(data)


def show_paradigm_application(title, word_rows, parad_application):
    context_dict = {"title": title, "word_rows": word_rows}
    parad_application_html = paradigm_application_template.render(context_dict)
    parad_application.value = parad_application_html


def show_paradigm(morpheme_res, words_inflections):
    parad_application = widgets.HTML(value="")
    word_rows = [
        ParadigmAppWordDesc(
            m.text, ", ".join(map(lambda f: f.feature_name, m.features))
        )
        for m in morpheme_res
    ]
    show_paradigm_application(
        "Paradigma Morfema Aplicado", word_rows, parad_application
    )
    messages = []
    if len(words_inflections) != 0:
        s_i = set(words_inflections)
        s_w = set([m.text for m in morpheme_res])
        c_words_inflections_expected_not_generated = list(s_i - s_w)
        c_words_inflections_generated_not_expected = list(s_w - s_i)
        if len(c_words_inflections_expected_not_generated) != 0:
            messages.append(
                "Inflexiones esperadas NO generadas: "
                + ", ".join(c_words_inflections_expected_not_generated)
            )
        if len(c_words_inflections_generated_not_expected) != 0:
            messages.append(
                "Inflexiones generadas NO esperadas: "
                + ", ".join(c_words_inflections_generated_not_expected)
            )
    if len(messages) != 0:
        message = "\n".join(messages)
        text_message = widgets.Textarea(
            value=message,
            description="Errores:",
            disabled=True,
            layout=widgets.Layout(width="350px", height="120px"),
        )
        parad_application = widgets.HBox([parad_application, text_message])
    # else:
    #     display(parad_application)

    return parad_application



import functools
import json

import mammut
from mammut.common.basic_tokenizer import BasicTokenizer
from mammut.common.corpus.constants import CorpusTable, CrawlerCorpusTable
from mammut.common.corpus.mammut_brat_document import MammutBratDocument
from mammut.common.lexicon.dictionary import Dictionary
from mammut.common.lexicon.mammut_brat_configuration import (
    MammutBratConfiguration,
)
from mammut.common.storage.storage_manager import StorageManager
from mammut.common.util import StringEnum, RegionalSettingsType
from mammut.common.corpus.corpus_operations import CorpusOperations
from typing import List


class ScenarioType(StringEnum):
    """
        Possible Scenario types.
    """
    Monologue = "Monologue"
    Conversation = "Conversation"
    Dialogue = "Dialogue"


class CorpusEventEntry:
    """This class encapsulates the concrete notion
    of a Corpus Event.

    An event is a message emitted by an agent in a channel.

    This class can be used as it is, or it can be extended by
    other specific events classes.

    TODO: Document class methods.
    """
    def __init__(
        self,
        sub_id: int,
        str_message: str,
        str_field: str,
        str_complexity: str,
        str_source: str,
        str_regional_settings: str,
        tokenizer: BasicTokenizer,
        dictionary: Dictionary,
        scenario_entry,
    ):
        self.valid = True
        self.errors = []
        self.tokenized_messages = []
        self.sub_id = sub_id
        self.field = str_field
        self.str_complexity = str_complexity
        self.str_source = str_source
        self.str_regional_settings = str_regional_settings
        self.tokenizer = tokenizer
        self.dictionary = dictionary
        self.scenario_entry = scenario_entry
        self.is_from_mammut = (
            CorpusTable.MAMMUT_USERS_PREFIX() in self.str_source.lower()
        )
        # Validating
        if self.str_complexity.isdigit():
            self.complexity = int(self.str_complexity)
        elif not self.str_complexity:
            self.complexity = 1
        else:
            self.complexity = 1
            self.valid = False
            self.errors.append(
                f"Invalid Complexity: not a digit. SubId = {self.sub_id}"
            )
        if RegionalSettingsType.contains_value(self.str_regional_settings):
            self.regional_settings = RegionalSettingsType(
                self.str_regional_settings
            )
        else:
            self.regional_settings = RegionalSettingsType.ES
            self.valid = False
            self.errors.append(
                f"Invalid Regional Settings: unknown regional settings {self.str_regional_settings}. SubId = {self.sub_id}"
            )
        self.add_message(str_message, False)

    def add_message(self, str_message: str, is_extension: bool):
        def add_error_of_invalid_message(error_msg, invalid_tokens):
            self.valid = False
            invalid_tokens_str = functools.reduce(
                lambda l, r: f"{l} - {r}", invalid_tokens, ""
            )
            if not is_extension:
                error_msg = (
                    "Invalid Meesage: "
                    + error_msg
                    + f". SubId = {self.sub_id}"
                )
            else:
                error_msg = (
                    "Invalid Extension Meesage: "
                    + error_msg
                    + f". SubId = {self.sub_id} - Message = {str_message}"
                )
            self.errors.append(
                error_msg + f" Invalid tokens: {invalid_tokens_str}"
            )

        tokenized_message = self.tokenizer.tokenize_corpus_entry(
            str_message,
            self.scenario_entry.corpus.annotable,
            self.str_regional_settings,
        )
        if tokenized_message.contains_invalid_variables():
            add_error_of_invalid_message(
                "the message contains invalid variables",
                tokenized_message.invalid_variables,
            )
        if tokenized_message.contains_invalid_images():
            add_error_of_invalid_message(
                "the message contains invalid images",
                tokenized_message.invalid_images,
            )
        if tokenized_message.contains_invalid_verbs():
            add_error_of_invalid_message(
                "the message contains invalid buttons",
                tokenized_message.invalid_verbs,
            )
        if tokenized_message.contains_unknown():
            add_error_of_invalid_message(
                "the message contains a token that is not qualifiy as valid",
                tokenized_message.unknowns,
            )
        self.tokenized_messages.append(tokenized_message)

    def get_annotation_documents(self, config: MammutBratConfiguration):
        return [
            MammutBratDocument(mt, config) for mt in self.tokenized_messages
        ]


class CorpusScenarioEntry:
    """This class encapsulates the general notion of a scenario.

    A scenario is a set of interactions that belongs to the same context,
    then they are a sequence of chronological corpus events. Each new sequence
    of events will be part of a new scenario.

    This class can be used as it is, or it can be extended by other
    specific scenario classes.

    # TODO: Document class methods
    """
    def __init__(
        self,
        str_id: str,
        str_type: str,
        scenerie_group_by_object,
        tokenizer: BasicTokenizer,
        dictionary: Dictionary,
        language_settings: str,
        corpus,
    ):
        self.valid = True
        self.contains_invalid_events = False
        self.errors = []
        self.str_id = str_id
        self.scenerie_group_by_object = scenerie_group_by_object
        self.tokenizer = tokenizer
        self.dictionary = dictionary
        self.language_settings = language_settings
        self.corpus = corpus
        self.str_type = str_type
        self.events = []
        # Global validations
        if self.str_id.isdigit():
            self.id = int(self.str_id)
        else:
            self.valid = False
            self.errors.append("Invalid Id: is not a integer")
        if self.str_type in list(ScenarioType.__members__):
            self.type = ScenarioType(self.str_type)
        else:
            self.type = ScenarioType.Conversation
            self.valid = False
            self.errors.append("Invalid Type: unknown scenario type")
        last_event_is_from_mammut = True
        last_event_source = ""
        index = 0
        all_hidden = True
        for e in self.scenerie_group_by_object.iterrows():
            if (
                CorpusTable.HIDDEN_COLUMN_NAME() not in e[1]
                or not e[1][CorpusTable.HIDDEN_COLUMN_NAME()]
            ):
                cee = self.get_corpus_event_entry_from_event_row(index, e)
                self.events.append(cee)
                self.validate_corpus_event_entry(
                    index, cee, last_event_source, last_event_is_from_mammut
                )
                last_event_source = cee.str_source
                last_event_is_from_mammut = cee.is_from_mammut
                index = index + 1
                all_hidden = False
        if (
            self.type is not ScenarioType.Monologue
            and len(self.events) < 2
            and not all_hidden
        ):
            self.valid = False
            self.errors.append(
                f"Invalid Amount of entries: less than 2. Id = {self.str_id}"
            )

    def get_corpus_event_entry_from_event_row(self, index, e):
        event_message = ""
        if (
            CorpusTable.EVENT_MESSAGE_COLUMN_NAME() in e[1]
            and e[1][CorpusTable.EVENT_MESSAGE_COLUMN_NAME()]
        ):
            event_message = e[1][CorpusTable.EVENT_MESSAGE_COLUMN_NAME()]

        return CorpusEventEntry(
            index,
            event_message,
            e[1][CorpusTable.FIELD_COLUMN_NAME()]
            if CorpusTable.FIELD_COLUMN_NAME() in e[1]
            else "",
            e[1][CorpusTable.COMPLEXITY_COLUMN_NAME()]
            if CorpusTable.COMPLEXITY_COLUMN_NAME() in e[1]
            else "",
            e[1][CorpusTable.SOURCE_COLUMN_NAME()]
            if CorpusTable.SOURCE_COLUMN_NAME() in e[1]
            else "",
            e[1][CorpusTable.REGIONAL_SETTINGS_COLUMN_NAME()],
            self.tokenizer,
            self.dictionary,
            self,
        )

    def validate_corpus_event_entry(
        self, index, cee, last_event_source, last_event_is_from_mammut
    ):
        # Validations at event level
        if not cee.valid and not self.contains_invalid_events:
            self.valid = False
            self.contains_invalid_events = True
            self.errors.append("Contains invalid events")
            self.errors.extend(cee.errors)
        if (
            self.type is not ScenarioType.Monologue
            and cee.str_source == last_event_source
        ):
            self.valid = False
            self.errors.append(
                f"Invalid sequence of events: two consecutive events from the same source. Id = {self.str_id} - SubId = {index}"
            )
        if (
            self.type is ScenarioType.Conversation
            and not last_event_is_from_mammut
            and not cee.is_from_mammut
        ):
            self.valid = False
            self.errors.append(
                f"Invalid sequence of events: two consecutive events from users in a conversation scenario. Id = {self.str_id} - SubId = {index}"
            )
        if self.type is ScenarioType.Monologue:
            if cee.is_from_mammut:
                self.valid = False
                self.errors.append(
                    f"Invalid sequence of events: one event from mammut in a monologue scenario. Id = {self.str_id} - SubId = {index}"
                )
            elif cee.str_source != last_event_source and index != 0:
                self.valid = False
                self.errors.append(
                    f"Invalid sequence of events: more than one source in a monologue scenario. Id = {self.str_id} - SubId = {index}"
                )

    def add_event_extension(self, sub_id: int, str_message: str):
        if sub_id < len(self.events):
            self.events[sub_id].add_message(str_message, True)
        else:
            self.valid = False
            self.errors.append(
                f"Invalid Extension Meesage: the message has a sub-id greater or equal to the amount of events in the scenario . SubId = {sub_id} - Message = {str_message}"
            )

    def add_event_extensions(self, scenerie_ext_group_by_object):
        for ee in scenerie_ext_group_by_object.iterrows():
            if (
                CorpusTable.HIDDEN_COLUMN_NAME() not in ee[1]
                or not ee[1][CorpusTable.HIDDEN_COLUMN_NAME()]
            ):
                sub_id_str = ee[1][CorpusTable.SUB_ID_COLUMN_NAME()]
                str_message = ""
                if CorpusTable.EVENT_MESSAGE_COLUMN_NAME() in ee[1]:
                    str_message = ee[1][
                        CorpusTable.EVENT_MESSAGE_COLUMN_NAME()
                    ]
                if sub_id_str.isdigit():
                    sub_id = int(sub_id_str)
                    self.add_event_extension(sub_id, str_message)
                else:
                    self.valid = False
                    self.errors.append(
                        f"Invalid Extension Meesage: the message has a invalid sub-id. SubId = {sub_id_str} - Message = {str_message}"
                    )


class Corpus(CorpusOperations):
    """This class abstracts the concept of Corpus.

    A Corpus is a set of prototypical conversations that represents
    the communication between different agents, typically a bot and one
    or more human users.

    TODO: document class methods.
    """
    DEFAULT_CORPUS_START_COLUMN = "A"
    DEFAULT_CORPUS_START_ROW = "2"
    DEFAULT_CORPUS_END_CELL = "N"
    DEFAULT_CORPUS_EXTENSION_END_CELL = "G"
    DEFAULT_CONFIGURATION_LOAD_DATA_INTERVAL = "load_data_interval"
    storage_manager = StorageManager()

    def __init__(
        self,
        spreadsheet_id: str,
        id,
        corpus_sheet_title: str,
        corpus_extension_sheet_title: str,
        tokenizer: BasicTokenizer,
        dictionary: Dictionary,
        regional_settings: str,
        annotable: bool,
        configuration_json=None,
        corpus_dataframe=None,
    ):
        self.valid = True
        self.contains_invalid_sceneries = False
        self.errors = []
        self.sceneries = {}
        self.invalid_sceneries = []
        self.spreadsheet_id = spreadsheet_id
        self.id = id
        self.corpus_sheet_title = corpus_sheet_title
        self.sheet_title = self.corpus_sheet_title
        self.corpus_extension_sheet_title = corpus_extension_sheet_title
        self.tokenizer = tokenizer
        self.dictionary = dictionary
        self.regional_settings = regional_settings
        self._annotable = annotable

        if configuration_json is not None:
            self.configuration = json.loads(configuration_json)
            temporal_sheets = self.storage_manager.get_temporal_sheets_from_new_to_old(
                self.spreadsheet_id, self.corpus_sheet_title
            )
            if (
                Corpus.DEFAULT_CONFIGURATION_LOAD_DATA_INTERVAL
                in self.configuration
            ):
                if (
                    len(temporal_sheets) == 0
                    or (
                        mammut.get_timezone_aware_datetime_now()
                        - temporal_sheets[0][2]
                    ).total_seconds()
                    > self.configuration[
                        Corpus.DEFAULT_CONFIGURATION_LOAD_DATA_INTERVAL
                    ]
                ):
                    self.dump_sheet_title = self.create_dump_sheet(
                        spreadsheet_id
                    )
                    self.corpus_sheet_title = self.dump_sheet_title
                    self.DEFAULT_CORPUS_RANGE = self.dump_sheet_title + "!A3"
                    self.load_data()
                    for ts in temporal_sheets:
                        self.storage_manager.delete_sheet(
                            spreadsheet_id, ts[0]
                        )
                else:
                    self.dump_sheet_title = temporal_sheets[0][1]
                    self.corpus_sheet_title = self.dump_sheet_title
                    self.DEFAULT_CORPUS_RANGE = self.dump_sheet_title + "!A3"

        if corpus_dataframe is not None:
            self.corpus_data_frame = corpus_dataframe
        else:
            self.corpus_data_frame = self.storage_manager.get_spreadsheet_as_dataframe(
                self.spreadsheet_id,
                self.corpus_sheet_title,
                self.DEFAULT_CORPUS_START_COLUMN,
                self.DEFAULT_CORPUS_START_ROW,
                self.DEFAULT_CORPUS_END_CELL,
            )
        self.sceneries_group_by_object = self.corpus_data_frame.groupby(
            CorpusTable.ID_COLUMN_NAME()
        )
        for g in self.sceneries_group_by_object:
            cse = self.get_corpus_scenario_entry_from_row(g)
            if len(cse.events) > 0:
                if cse.valid:
                    if cse.id not in self.sceneries:
                        self.sceneries[cse.id] = cse
                    else:
                        self.invalid_sceneries.append(cse)
                        self.valid = False
                        self.errors.append(
                            f"Invalid scenario ID: repeated id. Id = {cse.id}"
                        )
                else:
                    self.invalid_sceneries.append(cse)
                    self.valid = False
                    if not self.contains_invalid_sceneries:
                        self.contains_invalid_sceneries = True
                        self.errors.append("Contains invalid sceneries")
        if self.corpus_extension_sheet_title:
            self.corpus_ext_data_frame = self.storage_manager.get_spreadsheet_as_dataframe(
                self.spreadsheet_id,
                self.corpus_extension_sheet_title,
                self.DEFAULT_CORPUS_START_COLUMN,
                self.DEFAULT_CORPUS_START_ROW,
                self.DEFAULT_CORPUS_EXTENSION_END_CELL,
            )

            self.sceneries_ext_group_by_object = self.corpus_ext_data_frame.groupby(
                [CorpusTable.ID_COLUMN_NAME()]
            )
            for ge in self.sceneries_ext_group_by_object:
                str_id = ge[0]
                if str_id.isdigit():
                    id = int(str_id)
                    if id in self.sceneries:
                        self.sceneries[id].add_event_extensions(ge[1])
                        if not self.sceneries[id].valid:
                            self.valid = False
                            if (
                                self.sceneries[id]
                                not in self.invalid_sceneries
                            ):
                                self.invalid_sceneries.append(
                                    self.sceneries[id]
                                )
                            if not self.contains_invalid_sceneries:
                                self.contains_invalid_sceneries = True
                                self.errors.append(
                                    "Contains invalid sceneries"
                                )
                else:
                    if not self.contains_invalid_sceneries:
                        self.contains_invalid_sceneries = True
                        self.errors.append("Contains invalid sceneries")
                    self.valid = False
                    self.errors.append(
                        "Invalid Id: id in corpus extension is not a integer"
                    )

    def get_corpus_scenario_entry_from_row(self, g):
        raise NotImplementedError()

    def load_data(self):
        raise NotImplementedError()

    def create_dump_sheet(self, spreadsheet_id):
        temporal_sheet_title = self.storage_manager.create_temporal_new_sheet(
            spreadsheet_id, self.corpus_sheet_title
        )
        self.storage_manager.add_rows_to_spreadsheet(
            spreadsheet_id,
            temporal_sheet_title,
            CrawlerCorpusTable.DEFAULT_COLUMN_TITTLES,
            CrawlerCorpusTable.DEFAULT_START_ROW_TITTLE,
            CrawlerCorpusTable.DEFAULT_START_COLUMN_TITTLE,
        )
        return temporal_sheet_title

    def print_validation_result(self):

        output = []
        if self.valid:
            output.append("✔ El corpus es valido")
        else:
            output.append("✘ El corpus NO es valido")
            if len(self.errors) > 0:
                print(
                    f"Errores a nivel del corpus: ({len(self.errors)} errores)"
                )
                for e in self.errors:
                    print("\t" + e)
            if self.contains_invalid_sceneries:
                print("Errores a nivel de los escenarios:")
                for s in self.invalid_sceneries:
                    for es in s.errors:
                        print("\t" + es)
                    if s.contains_invalid_events:
                        print(
                            f"\tErrores a nivel de los eventos del escenario {s.str_id}:"
                        )
                        for ev in s.events:
                            if not ev.valid:
                                for eev in ev.errors:
                                    print("\t\t" + eev)

        return output

    @property
    def indexable(self) -> bool:
        return True

    @property
    def annotable(self) -> bool:
        return self._annotable

    def get_events_as_strings(self) -> List[str]:
        """
            Not implemented here, will require deep work.
        """
        raise NotImplementedError()





# coding=utf-8
import urllib.request as Request
from urllib.parse import urlparse
import urllib
from lxml import html, etree
import time


class CrawlingProperties:
    def __init__(self, base_url, nodes_pattern, node_elements, page_href = '', delay = 2, pagination_nodes_pattern = '',
                 pagination_element_xpath = '', pages_limit: int=None):
        self.base_url = base_url
        self.page_href = page_href
        self.nodes_pattern = nodes_pattern
        self.node_elements = node_elements
        self.delay = delay
        self.pagination_nodes_pattern = pagination_nodes_pattern
        self.pagination_element_xpath = pagination_element_xpath
        self.pages_limit = pages_limit

class MammutCrawler:
    """
    mammut crawler class
    """
    def __init__(self, crawling_properties: CrawlingProperties, update_status_text_box = None,  mammut_crawler = None):
        self.crawling_properties = crawling_properties
        self.update_status_text_box = update_status_text_box
        self.mammut_crawler = mammut_crawler

    def get_page(self, url):
        """
        return page string
        """
        req = Request.Request(url, headers={'User-Agent': 'Mozilla/5.0'})
        request_response = Request.urlopen(req)
        page_string = request_response.read()
        return page_string

    def parse_string(self, page_tring):
        """
        lxml parsed string
        """
        page_lxml = html.fromstring(page_tring)
        return page_lxml

    def get_nodes(self, page_lxml, pattern):
        """
        elements that match pattern, return eTree objects as lists
        """
        nodes = page_lxml.xpath(pattern)
        return nodes

    def get_nodes_info(self, nodes, node_elements):
        """
        return information about the nodes like url,title or other to continue with crawling actions
        """
        nodes_info = []
        for node in nodes:
            nodes_info.append(self.get_node_data(node, node_elements))
        return nodes_info

    def get_node_data(self, node, node_elements):
        """
        nodeElemets serian un array de xpath relativos de las partes del articulo,receta,documento o lo que sea que se revise
        returns data of specified elements(xpaths) from one node
        """
        node_data = {}
        for elem in node_elements:
            node_data[elem] = self.get_node_element_data(node, node_elements[elem])
        return node_data

    def get_node_element_data(self, node, element_relative_path):
        """
        return data of a node, it need a relative path to the node to get the relevant data like: title, image, url, body...
        """
        data = ""
        try:
            if type(element_relative_path) == str:
                elem = node.xpath(element_relative_path)
                for i, j in enumerate(elem):
                    data = data + elem[i]
            else:
                data = element_relative_path
        except etree.XPathError:
            data = element_relative_path
        if data == '' and not element_relative_path.startswith('/'):
            data = element_relative_path
        return data

    def get_page_info(self, page_string, nodes_pattern, node_elements):
        parsed_page = self.parse_string(page_string)
        nodes = self.get_nodes(parsed_page, nodes_pattern)
        nodes_info = self.get_nodes_info(nodes, node_elements)
        return nodes_info

    def get_next_href(self, page_string, pagination_nodes_pattern, pagination_element_xpath):
        parsed_page = self.parse_string(page_string)
        pagination_nodes = self.get_nodes(parsed_page, pagination_nodes_pattern)
        if len(pagination_nodes) != 0:
            pagination_href_element = self.get_node_element_data(pagination_nodes[-1], pagination_element_xpath)
        else:
            pagination_href_element = ''
        if pagination_href_element != pagination_element_xpath and pagination_href_element[0] != '/':
            pagination_href_element = '/' + pagination_href_element

        return pagination_href_element

    def get_pages_info(self, test=False):
        next_href = self.crawling_properties.page_href
        all_nodes = []
        count = 0
        while True:
            time.sleep(self.crawling_properties.delay)
            # if next_href == self.crawling_properties.page_href:
            #     full_url = self.crawling_properties.base_url + next_href
            if next_href[0] == '/':
                url_path = urlparse(next_href[1:])
            else:
                url_path = urlparse(next_href)
            if url_path.netloc:
                full_url = url_path.geturl()
            else:
                #TODO: revisar esta logica del crawler, no funciona para bajar el corpus de degusta.
                # if next_href == self.crawling_properties.page_href:
                #     joined_path = urllib.parse.urljoin(urlparse(full_url).path, urlparse(next_href).path)
                # elif next_href[0] == '/':
                #     joined_path = urllib.parse.urljoin(urlparse(full_url).path, urlparse(next_href[1:]).path)
                # full_url = urllib.parse.urljoin(full_url, joined_path)
                full_url = urllib.parse.urljoin(self.crawling_properties.base_url, url_path.path)
            if self.update_status_text_box is not None:
                self.update_status_text_box()
            page_string = self.get_page(full_url)
            nodes_info = self.get_page_info(page_string, self.crawling_properties.nodes_pattern,
                                            self.crawling_properties.node_elements)
            for i in nodes_info:
                all_nodes.append(i)
            if self.crawling_properties.pagination_nodes_pattern != '' and self.crawling_properties.pagination_element_xpath != '':
                next_href = self.get_next_href(page_string, self.crawling_properties.pagination_nodes_pattern,
                                               self.crawling_properties.pagination_element_xpath)
            else:
                next_href = ''
            if not next_href.startswith('/'):#TODO: esta condicion garantiza que no es un url???
                break
            if test:
                break
            if self.crawling_properties.pages_limit - 1 == count:
                break
            count += 1
        return all_nodes

    def execute(self, test=False):
        all_of_all_data = []
        current_level_crawling_results = self.get_pages_info(test=test)
        if self.mammut_crawler == None:
            return current_level_crawling_results
        else:
            if test:
                current_level_crawling_results = current_level_crawling_results[:2]
            for index, i in enumerate(current_level_crawling_results):
                self.mammut_crawler.crawling_properties.page_href = i[list(i)[0]]
                page_info = self.mammut_crawler.execute(test=test)
                all_of_all_data = all_of_all_data + page_info
        return all_of_all_data
